/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2005
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{ed857dc2-1d44-11da-94ca-00e08161165f}");
const CLASS_NAME = "Checkyesss Container JS Component";
const CONTRACT_ID = "@mozdev.org/checkyesss/container;1";


/***************************************************************************/

/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function StringArrayEnumerator(array) {
    this.array = array;
    this.index = 0;
}
StringArrayEnumerator.prototype = {
    hasMore: function() { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                                  this.array[this.index++] : null; }
};

/***************************************************************************/

/**
/* The Checkyesss Container Class
 */

function CheckyesssContainerClass() {
  //this.wrappedJSObject = this;
  this._data = new Object();
}


CheckyesssContainerClass.prototype = {
  
  setPropertyAsString: function(name, value) {
    if (typeof(value) != "string")
      value = "" + value;
    this._data[" "+name] = value;
  },
  
  
  setPropertyAsBool: function(name, value) {
    if (typeof(value) != "boolean")
      value = (value) ? true : false;
    this._data[" "+name] = value;
  },
  
  
  setPropertyAsInt: function(name, value) {
    if (typeof(value) != "number")
      value = parseInt(value, 10);
    this._data[" "+name] = value;
  },
  
  
  setPropertyAsISupports: function(name, value) {
    if ((typeof(value) == "object") &&
        (value != null))
      this._data[" "+name] = value.QueryInterface(Components.interfaces.nsISupports);
    else
      this._data[" "+name] = null;
  },
  
  
  setPropertyFromAttribute: function(name, element) {
    if (element && element.hasAttribute(name))
      this._data[" "+name] = element.getAttribute(name);
  },
  
  
  setPropertiesFromAttributes: function(element) {
    if (element && element.attributes) {
      var attr = element.attributes;
      var i;
      for (i = 0; i < attr.length; i++)
        this._data[" "+attr.item(i).nodeName] = attr.item(i).nodeValue;
    }
  },
  
  
  getPropertyType: function(name) {
    if ((" "+name) in this._data)
      return typeof(this._data[" "+name]);
    else
      return "";
  },
  
  
  getPropertyAsString: function(name) {
    var val = null;
    if ((" "+name) in this._data) {
      val = this._data[" "+name];
      switch (typeof(val)) {
        case "number":
        case "boolean":
          val = ""+val;
          break;
        case "string":
          // just keep the value
          break;
        default:
          val = null;
          break;
      }
    }
    return val;
  },
  
  
  getPropertyAsBool: function(name) {
    var val = false;
    if ((" "+name) in this._data) {
      val = this._data[" "+name];
      switch (typeof(val)) {
        case "string":
          val = (val == "true" || val == "1");
          break;
        case "number":
          val = (val == 1);
          break;
        case "boolean":
          // just keep the value
          break;
        default:
          val = false;
          break;
      }
    }
    return val;
  },
  
  
  getPropertyAsInt: function(name) {
    var val = 0;
    if ((" "+name) in this._data) {
      val = this._data[" "+name];
      if (typeof(val) == "string") {
        /*jsl:ignore*/
        // no radix specified to allow hex/octal numbers
        val = parseInt(val);
        /*jsl:end*/
      }
      else if (typeof(val) != "number")
        val = 0;
    }
    return val;
  },
  
  
  getPropertyAsISupports: function(name) {
    var val = null;
    if ((" "+name) in this._data) {
      val = this._data[" "+name];
      if (typeof(val) == "object") {
        try {
          val = val.QueryInterface(Components.interfaces.nsISupports);
        }
        catch (e) {
          val = null;
        }
      }
      else
        val = null;
    }
    return val;
  },
  
  
  getPropertyAsAttribute: function(name, element) {
    if ((" "+name) in this._data) {
      var val = this._data[" "+name];
      var type = typeof(val);
      if (type == "string" || type == "number" || type == "boolean")
        element.setAttribute(name, val);
    }
    return element;
  },
  
  
  getPropertiesAsAttributes: function(element) {
    if (!element)
      return null;
    
    var val, type, n;
    for (n in this._data) {
      val = this._data[n];
      type = typeof(val);
      if (type == "string" || type == "number" || type == "boolean")
        element.setAttribute(n.substr(1), ""+val);
    }
    return element;
  },
  
  
  deleteProperty: function(name) {
    if ((" "+name) in this._data)
      delete this._data[" "+name];
  },
  
  
  // returns true if at least one property was removed
  keepProperties: function(names) {
    var filter = new Object();
    var removed = false;
    var n;
    
    while (names.hasMore())
      filter[" " + names.getNext()] = 1;
    
    for (n in this._data) {
      if (!(n in filter)) {
        delete this._data[n];
        removed = true;
      }
    }
    
    return removed;
  },
  
  
  hasProperty: function(name) {
    return ((" "+name) in this._data);
  },
  
  
  getPropertyNames: function() {
    var names = new Array();
    var index = 0;
    var n;
    for (n in this._data) {
      if (n.charAt(0) == " ")
        names[index++] = n.substr(1);
    }
    return new StringArrayEnumerator(names);
  },
  
  
  clone: function() {
    // can handle simple data types only
    var cpy = Components.classes["@mozdev.org/checkyesss/container;1"].
                createInstance(Components.interfaces.cyICheckyesssContainer);
    var type, name, n;
    
    for (n in this._data) {
      type = typeof(this._data[n]);
      name = n.substr(1);
      switch (type) {
        case "string":
          cpy.setPropertyAsString(name, this._data[n]);
          break;
        case "number":
          cpy.setPropertyAsInt(name, this._data[n]);
          break;
        case "boolean":
          cpy.setPropertyAsBool(name, this._data[n]);
          break;
        default:
          break;
      }
    }
    return cpy;
  },
  
  
  isEquivalent: function(aCyContainer, aCheckProps) {
    var equiv = true;
    var prop;
    
    if (aCheckProps) {
      var exist1, exist2;
      while (equiv && aCheckProps.hasMore()) {
        prop = aCheckProps.getNext();
        exist1 = (" "+prop) in this._data;
        exist2 = aCyContainer.hasProperty(prop);
        equiv = equiv && (exist1 == exist2) &&
                         (!exist1 || (this._data[prop] ==
                                      aCyContainer.getPropertyAsString(prop)));
      // TODO: do not just compare as strings but according to our property's type!!!
      }
    }
    else {
      var propStrip;
      // first compare our own properties
      for (prop in this._data) {
        if (!prop.charAt(0) == " ")          // some strange property
          continue;
        propStrip = prop.substr(1);
        if (!aCyContainer.hasProperty(propStrip) ||
            (this._data[prop] != aCyContainer.getPropertyAsString(propStrip)))
          return false;
      }
      // now check if the other container has properties which we don't
      var props = aCyContainer.getPropertyNames();
      while (props.hasMore()) {
        if (!((" " + props.getNext()) in this._data))
          return false;
      }
    }
    return equiv;
  },
  
  
  empty: function() {
    /* really empties the object instead of creating an new empty object!! */
    var n;
    for (n in this._data)
      delete this._data[n];
  },
  
  
  get isEmpty() {
    var n;
    for (n in this._data)
      return false;
    return true;
  },
  
  
  toString: function() {
    var str = "";
    var n, type, name;
    for (n in this._data) {
      type = typeof(this._data[n]);
      name = n.substr(1);
      str += name;
      if (type == "string" || type == "number" || type == "boolean")
        str += "("+this._data[n]+") ";
      else
        str += "(??) ";
    }
    return str;
  },
  
  
  /**********************************************************************/
  /*          X P C O M - S P E C I F I C   M E T H O D S               */
  /**********************************************************************/

  QueryInterface: function (aIID) {
    if (!aIID.equals(Components.interfaces.cyICheckyesssContainer) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/

var gModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID,
                                     CLASS_NAME,
                                     CONTRACT_ID,
                                     aFileSpec,
                                     aLocation,
                                     aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID) {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) {
    return true;
  }
};

var gFactory = {
  
  createInstance: function (aOuter, aIID) {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (!aIID.equals(Components.interfaces.cyICheckyesssContainer) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    
    return ((new CheckyesssContainerClass()).QueryInterface(aIID));
  }
};

function NSGetModule(aCompMgr, aFileSpec) {
  return gModule;
}

