/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2005-2008
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{f134ea90-cc83-11da-a94d-0800200c9a66}");
const CLASS_NAME = "Checkyesss sipgate-de Manager Component";
const CONTRACT_ID = "@mozdev.org/checkyesss/sipgate_manager;2";

const TO_HTTP = 120;              // 120 seconds until aborting stalled access
const TO_MIN_PAUSE = 10;          // 10 minutes pause between scheduled accesses
const TO_BLOCKED = 60;            // 60 seconds pause if access was blocked
const TO_MIN_CALLUPD = 1;         // min. pause for calls, 0..update immediatly
const TO_CALLUPD = 7*24;          // 7 days regular interval for calls

const MAX_CARDS = 10;

const DEBUG = 0;


/***************************************************************************/

/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function ArrayEnumerator(array) {
    this.array = array;
    this.index = 0;
}
ArrayEnumerator.prototype = {
    hasMoreElements: function () { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};
// nsISupportsArrayEnumerator
//
function nsISupportsArrayEnumerator(array) {
    this.array = array;
    this.index = 0;
}
nsISupportsArrayEnumerator.prototype = {
    hasMoreElements: function() { return this.index < this.array.Count(); },
    getNext: function () { return (this.index < this.array.Count()) ?
                       this.array.GetElementAt(this.index++) : null; }
};

/**
/* StringEnumerator
 */
function StringEnumerator(array) {
    this.array = array;
    this.index = 0;
}
StringEnumerator.prototype = {
    hasMore: function() { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};

/***************************************************************************/

/**
/* 
 */
function AccountDetails(aId, aUser, aPassword, aPlan, aEnabled, aMayStorePwd) {
  this._id = aId;
  this._user = aUser;            // user name
  this._pw = ((typeof(aPassword) == "string") && (aPassword.length > 0)) ?
               aPassword : null;          // password
  this._mayStorePwd = aMayStorePwd;   // may the password be stored in PW-Mng?
  this._en = aEnabled;           // account enabled
  this._rateplan = aPlan;        // selected plan
  this._phone = "??";            // phone number
  this._credit = "";             // current credit ("euro.cent")
  this._messages = 0;            // number of new messages
  this._dateCallsLoaded = 0;     // last time calls where updated
  this._error = "Daten noch nicht geladen";
  this._forceupdate = false;     // force update!!
  this._quickLinkPending = false;
  this._quickLinkWindow = null;
  this._quickLinkTarget = "";
  this._reloadAvailableCalls = false;
}

/******************************************************************************
 * CheckyesssSipgateManager component
 ******************************************************************************/

function CheckyesssSipgateManager() {
  this.wrappedJSObject = this;
  this._start();
}


CheckyesssSipgateManager.prototype = {
  
  _providerId:   "SG002",
  
  _sipgateLogin:  "https://secure.sipgate.de/user/index.php",
  _sipgateCredit: "https://secure.sipgate.de/user/kontostand.php",
  _sipgateDeposits: "https://secure.sipgate.de/user/kontostand.php?show=all",
  _sipgateCalls:  "https://secure.sipgate.de/user/konto_einzel.php?show=all",
  _sipgateLogout: "https://secure.sipgate.de/user/logout.php",
  
  _blockingUrlRegExp: "^https://secure\\.sipgate\\.de/user/",
  
  _timeHour:   3600000,
  _timeMinute: 60000,
  _timeSecond: 1000,
  
  _observerTopic: null,
  _userRegExp: null,
  
  _interval: 24,                     // interval [hours]
  _suspendUntil: 0,                  // suspend automatic update from internet
  
  _accountsArray: null,
  _accountsChanged: false,
  _accountsTransferred: null,
  
  _webAccess: null,
  _accessAccount: -1,
  _lastAccess: 0,     // date of last access [ms] (not counting logout)
  _checkPending: false,
  
  _firstTime: true,
  
  _checkTimer: null,
  _obs: null,
  _utils: null,
  _userAgent: null,
  _managerCenter: null,
  
  
  /**************************************************************/
  
  _start: function() {
    // set components
    this._utils = Components.classes["@mozdev.org/checkyesss/util_service;1"].
                getService(Components.interfaces.cyICheckyesssUtil);
    
    this._checkTimer = Components.classes["@mozilla.org/timer;1"].
                         createInstance(Components.interfaces.nsITimer); 
    
    this._obs = Components.classes["@mozilla.org/observer-service;1"].
                  getService(Components.interfaces.nsIObserverService);
    
    this._managerCenter = Components.classes["@mozdev.org/checkyesss/manager_center;1"].
                            getService(Components.interfaces.cyICheckyesssCenter);
    
    var providerProperties = this._managerCenter.getProviderProperties(this._providerId);
    this._observerTopic = providerProperties.getPropertyAsString("topic");
    this._userRegExp = providerProperties.getPropertyAsString("validUserRegExp");
    
    //add observer topics
    this._obs.addObserver(this, "xpcom-shutdown", false);
    
    this._webAccess = Components.classes["@mozdev.org/checkyesss/web_access;1"].
                        createInstance(Components.interfaces.cyICheckyesssWeb);
    this._webAccess.requestTimeout = TO_HTTP;
    
    try {
      this._userAgent = Components.classes["@mozilla.org/network/protocol;1?name=http"].
                          getService(Components.interfaces.nsIHttpProtocolHandler).
                            userAgent;
      this._userAgent += " Checkyesss/0.6";
    }
    catch (e) {
      this._userAgent = null;
    }
    
    if (DEBUG) {
      this._sipgateLogin  = "http://localhost/sg/start.php.html";
      //this._sipgateLogin  = "http://localhost/sg/fehler.php.html";
      this._sipgateLogout = "http://localhost/sg/logout.html";
      this._sipgateCalls  = "http://localhost/sg/konto_einzel-de.php.html";
      this._sipgateDeposits = "http://localhost/sg/kontostand-zeit.php.html";
      this._sipgateCredit = "http://localhost/sg/kontostand-zeit.php.html";
      this._timeHour = this._timeMinute;
      this._timeMinute = this._timeSecond;
    }
    
    //this._checkTimer.initWithCallback(this, 10, 
    //                                  Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "Checkyesss SIPGATE-Manager started", "", false);
    
    this._firstTime = false;
  },
  
  /**************************************************************/
  
  _stop: function() {
    this._webAccess.endSession();
    
    // stop timers
    this._checkTimer.cancel();
    
    // clean up observers
    try {
      this._obs.removeObserver(this, "xpcom-shutdown");
    }
    catch(e) {}
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "Checkyesss SIPGATE-Manager stopped", "", false);
    
    //destroy variables
    this._webAccess = null;
    //this._busy = false;
    this._jumpWin = null;
    this._checkTimer = null;
    this._errtxt = null;
    //this._loginCallback = null;
    this._firstTime = true;
    this._accountsArray = null;
    this._utils = null;
    this._managerCenter = null;
  },
  
  /**************************************************************/
  
  get updateInterval() {
    return this._interval;
  },
  
  set updateInterval(aValue) {
    this._interval = (aValue <  4) ? 4 :
                     (aValue > 24) ? 24 :
                     aValue;
    return aValue;
  },
  
  /**************************************************************/
  
  get suspendNetAccessUntil() {
    return this._suspendUntil;
  },
  
  set suspendNetAccessUntil(aValue) {
    this._suspendUntil = aValue;
    return aValue;
  },
  
  /**************************************************************/
  
  updateClients: function() {
    this._refreshDisplay();
  },
  
  /**************************************************************/
  
  clearAccounts: function() {
    if (!this._accountsArray ||
        (this._accountsArray.length == 0))
      return;
    
    this.abortAccess();
    
    this._accountsArray = new Array();
    
    this._obs.notifyObservers(this, this._observerTopic, "reset_all");
  },

  /**************************************************************/
  
  addAccount: function(aAccountId, aUser, aPassword, aRatePlanId,
                       aPeriodReset, isEnabled, aMayStorePwd) {
    var result = true;
    
    if (!this._accountsArray)
      this._accountsArray = new Array();
    var index = this._accountsArray.length;
    
    var entry = new AccountDetails(aAccountId, aUser, aPassword, aRatePlanId,
                                   isEnabled, aMayStorePwd);
    
    var ratePlanOk = false;
    var ratePlanIds = this._managerCenter.getRatePlanList(this._providerId);
    while (!ratePlanOk && ratePlanIds.hasMore())
      ratePlanOk = (ratePlanIds.getNext() == aRatePlanId);
    
    if (!ratePlanOk) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssSipgateDE.addAccount",
                                "Invalid rate plan ID '" + aRatePlanId + "'",
                                false);
      entry._en = false;
      result = false;
    }
    
    this._accountsArray[index] = entry;
    
    if (entry._en) {
      this._updateFromCache(index);
      
      // restart timer with a short timeout
      this._checkTimer.cancel();
      this._checkTimer.initWithCallback(this, 100, 
                                        Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }
    return result;
  },
  
  /**************************************************************/
  
  _assembleCreditInfo: function(aIndex, aDocument) {
    var entry;
    try {
      entry = this._accountsArray[aIndex];
    }
    catch (e) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "_assembleCreditInfo",
                                "Invalid internal index " + aIndex, false);
      return null;
    }
    
    if (!entry._en)
      return null;
    
    var expired = true;
    var age = -1;
    if (this._utils.existCache(entry._id)) {
      age = this._utils.getCacheAge(entry._id);
      expired = (this._utils.getCacheValidity(entry._id) == 0);
    }
    
    var container = Components.classes["@mozdev.org/checkyesss/container;1"].
                      createInstance(Components.interfaces.cyICheckyesssContainer);
    
    var tooltip = null;
    var isError = (entry._error || (entry._credit.toString().length == 0)) ? true : false;
    var pnum = entry._phone;
    
    var matches = pnum.match(/^\s*(0049|\+49|0)([0-9]+)\s*$/);
    if (matches)
      pnum = "0" + matches[2];
    
    if (aDocument) {
      // the following should be done with some transformation really...
      var descr, ltext;
      
      tooltip = aDocument.createDocumentFragment();
      
      if (entry._error) {
        var lines = entry._error.split("\n");
        for (var i = 0; i < lines.length; i++) {
          descr = aDocument.createElement("description");
          descr.setAttribute("class", "checkyesss-class-ttval");
          ltext = aDocument.createTextNode(lines[i]);
          descr.appendChild(ltext);
          tooltip.appendChild(descr);
        }
      }
      else if (entry._credit.toString().length == 0) {
        descr = aDocument.createElement("description");
        ltext = aDocument.createTextNode("Keine Daten vorhanden.");
        descr.appendChild(ltext);
        tooltip.appendChild(descr);
      }
      else {
        // credit: compute number of minutes and sms
        var credit = parseFloat(entry._credit);
        var min_home = parseInt(credit / 0.0179, 10);
        var min_cell = parseInt(credit / 0.1690, 10);
        var sms = parseInt(credit / 0.099, 10);
        var hours_home = this._utils.minutesToString(min_home) + " Festnetz";
        var hours_cell = this._utils.minutesToString(min_cell) + " Handy";
        sms = sms + " SMS";
        
        var node = aDocument.createElement("label");
        node.setAttribute("value", "Guthaben entspricht");
        node.setAttribute("class", "checkyesss-class-ttlab");
        tooltip.appendChild(node);
        
        node = aDocument.createElement("label");
        node.setAttribute("value", hours_home);
        node.setAttribute("class", "checkyesss-class-ttval");
        node.setAttribute("style", "margin-left: 1.5em;");
        tooltip.appendChild(node);
        
        node = aDocument.createElement("label");
        node.setAttribute("value", hours_cell);
        node.setAttribute("class", "checkyesss-class-ttval");
        node.setAttribute("style", "margin-left: 1.5em;");
        tooltip.appendChild(node);
        
        node = aDocument.createElement("label");
        node.setAttribute("value", sms);
        node.setAttribute("class", "checkyesss-class-ttval checkyesss-class-ttboxhdr");
        node.setAttribute("style", "margin-left: 1.5em;");
        tooltip.appendChild(node);
        
        if (entry._messages > 0) {
          node = aDocument.createElement("label");
          var str = "" + entry._messages + " neue Nachricht";
          if (entry._messages > 1)
            str += "en";
          node.setAttribute("value", str);
          node.setAttribute("class", "checkyesss-class-ttlab");
          tooltip.appendChild(node);
        }
      }
    }
    
    container.setPropertyAsString("userName", entry._user);
    container.setPropertyAsString("phoneNumber", pnum);
    container.setPropertyAsString("phoneCredit", entry._credit.toString());
    container.setPropertyAsInt("dataAge", age);
    container.setPropertyAsBool("dataExpired", expired);
    container.setPropertyAsBool("dataError", isError);
    container.setPropertyAsISupports("docTooltip", tooltip);
    
    return container;
  },

  /**************************************************************/
  
  getCreditInfo: function(aId, aDocument) {
    if (this._accountsArray) {
      for (var i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId)
          return this._assembleCreditInfo(i, aDocument);
      }
    }
    return null;
  },
  
  /**************************************************************/
  
  setManualAccountingStart: function(aAccountId, aStartTime) {
    null;
  },
  
  /**************************************************************/
  
  detectedRatePlan: function(aAccountId) {
    return "auto";
  },
  
  /**************************************************************/
  
  _refreshDisplay: function() {
    if (this._accountsArray) {
      for (var i = 0; i < this._accountsArray.length; i++) {
        entry = this._accountsArray[i];
        if (entry._en)
          this._obs.notifyObservers(this, this._observerTopic,
                                    "update::" + entry._id);
      }
    }
  },
  
  /**************************************************************/
  
  reloadCredit: function(aId) {
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId) {
          if (this._accountsArray[i]._en)
            index = i;
          break;
        }
      }
    }
    
    if (index >= 0) {
      var win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
      if (win) {
        var ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                   getService(Components.interfaces.nsIPromptService);    
        ps.alert(null, "Achtung",
                 "Das Guthaben kann derzeit nicht aktualisiert werden,\n"+
                 "da gerade ein sipgate-Konto angezeigt wird!\n\n"+
                 "Die Aktion wird verz\u00f6gert, bis alle Kontoseiten\n"+
                 "geschlossen wurden.");
        
        // focus the mentioned window
        //win.focus();
        
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesssSipgateDE.reloadCredit",
                                  "Conflicting browser window detected, action deferred",
                                  false);
      }
      
      this._accountsArray[index]._forceupdate = true;
      this._checkPending = true;
      
      // restart timer with a short timeout
      this._checkTimer.cancel();
      this._checkTimer.initWithCallback(this, 100, 
                                        Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }
  },
  
  /**************************************************************/
  
  updateCalls: function(aId, aForced) {
    var result = false;
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId) {
          if (this._accountsArray[i]._en)
            index = i;
          break;
        }
      }
    }
    
    if (index >= 0) {
      var win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
      if (win) {
        if (aForced) {
          var ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                     getService(Components.interfaces.nsIPromptService);    
          ps.alert(null, "Achtung",
                   "Die Einzelgespr\u00E4che k\u00f6nnen derzeit nicht aktualisiert\n"+
                   "werden, da gerade ein sipgate-Konto angezeigt wird!\n\n"+
                   "Bitte zuerst ausloggen und dann Aktion wiederholen!");
          
          // focus the mentioned window
          //win.focus();
          
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesssSipgateDE.updateCalls",
                                    "Conflicting browser window detected, action aborted",
                                    false);
        }
      }
      else {
        var entry = this._accountsArray[index];
        var now = (new Date()).valueOf();
        if (aForced ||
            (entry._dateCallsLoaded + (TO_MIN_CALLUPD * this._timeHour) < now)) {
          entry._forceupdate = true;
          entry._dateCallsLoaded = 0;
          this._checkPending = true;
          
          /* We need to circumvent the timer here as we have to make sure
             that the access is performed and a notification is returned!
             The usual 100ms timeout would open a time windows where
             conditions could change unexpectedly! */
          this._checkAccounts();
          
          result = true;      // expect a notification
        }
      }
    }
    return result;
  },
  
  /**************************************************************/
  
  getCallsRecords: function(aKey, aFromTime, aToTime) {
    var sum_credit = 0.0;
    var sum_fee = 0.0;         // fee for calls + sms
    var sum_cfee = 0.0;        // fee for all calls
    var sum_duration = 0;
    var sum_charged = 0;       // duration charged for
    var callsArray = new Array();
    var call, duration, fee, credit, charged, type, number;
    
    var calls = this._utils.getCallsRecords(aKey, aFromTime, aToTime, null);
    
    while (calls.hasMoreElements()) {
      call = calls.getNext().QueryInterface(Components.interfaces.cyICheckyesssContainer);
      
      number   = call.hasProperty("number") ? call.getPropertyAsString("number") : "";
      duration = call.hasProperty("duration") ? call.getPropertyAsInt("duration") : 0;
      fee      = call.hasProperty("fee") ? call.getPropertyAsString("fee") : "0";
      credit   = call.hasProperty("credit") ? call.getPropertyAsString("credit") : "0";
      type     = call.hasProperty("type") ? call.getPropertyAsString("type") : "";
      
      // replace Austrian country code by "0" to make numbers more readable
      number = number.replace(/^\+49/, "0");
      call.setPropertyAsString("number", number);
      
      fee    = Number(fee.replace(/[\.,]/, "."));
      credit = Number(credit.replace(/[\.,]/, "."));
      
      sum_duration += duration;
      sum_fee      += fee;
      sum_credit   += credit;
      
      if (fee >= 0 && type == "call") {
        // Now try to find the duration that's actually charged for.
        // THIS STRONGLY DEPENDS ON THE CURRENTLY AVAILABLE RATE PLANS!!!
        // Sipgate charges in increments of 60s.
        charged = duration - ((duration + 59) % 60) + 59;
        if (charged <= 60)          // special handling of 0 sec
          charged = 60;
        call.setPropertyAsInt("tcharged", charged);
        sum_charged += charged;
        sum_cfee += fee;
      }
      call.setPropertyAsString("stype", (type == "call") ? "Telefonat" :
                                        (type == "sms")  ? "SMS":
                                        (type == "connect")  ? "Verbindung":
                                        (type == "reload") ? "Aufladung" : "??");
      if (type != "call")
        call.deleteProperty("duration");
      
      if (fee >= 0 && !(credit > 0)) {
        fee = Math.round(fee * 10000.0);        // round to 1/100 cent
        fee = "00000" + fee.toString();
        fee = fee.replace(/^0*(\d+?)(\d{4})$/, "$1,$2");
        call.setPropertyAsString("fee", fee);
      }
      else if (call.hasProperty("fee"))
        call.deleteProperty("fee");
      
      if (credit > 0) {
        credit = credit.toString();
        call.setPropertyAsString("credit", credit);
      }
      else if (call.hasProperty("credit"))
        call.deleteProperty("credit");
      
      callsArray.push(call);
    }
    
    if (callsArray.length > 0) {
      call = Components.classes["@mozdev.org/checkyesss/container;1"].
               createInstance(Components.interfaces.cyICheckyesssContainer);
      call.setPropertyAsBool("statistic", true);
      call.setPropertyAsString("description", "Summe:");
      
      if (sum_duration > 0) {
        call.setPropertyAsInt("duration", sum_duration);
        call.setPropertyAsString("tcharged", this._utils.durationToString(sum_charged));
      }
      if (sum_fee > 0) {
        fee = Math.round(sum_fee * 10000.0);        // round to 1/100 cent
        fee = "00000" + fee.toString();
        fee = fee.replace(/^0*(\d+?)(\d{4})$/, "$1,$2");
        fee = "\u20AC " + fee;
        call.setPropertyAsString("fee", fee);
      }
      if (sum_credit > 0) {
        credit = "\u20AC " + sum_credit.toString();
        call.setPropertyAsString("credit", credit);
      }
      callsArray.push(call);
      
      if (sum_duration > 0 && sum_cfee > 0) {
        call = Components.classes["@mozdev.org/checkyesss/container;1"].
                 createInstance(Components.interfaces.cyICheckyesssContainer);
        call.setPropertyAsBool("statistic", true);
        call.setPropertyAsString("description", "Durchschnitt:");
        var average = (sum_cfee * 600000.0) / sum_duration;
        average = Math.round(average);                 // round to 1/100 cent
        fee = "00" + average.toString();
        fee = fee.replace(/^0*(\d+?)(\d{2})$/, "$1,$2");
        fee = fee + " \u00A2/min";                       // cent per minute
        call.setPropertyAsString("fee", fee);
        callsArray.push(call);
      }
    }
    
    return new ArrayEnumerator(callsArray);
  },
  
  /**************************************************************/
  
  enableReloadAvailableCalls: function(aAccountId) {
    //dump("enableReloadAvailableCalls\n");
    var entry = null;
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aAccountId) {
          entry = this._accountsArray[i];
          index = i;
          break;
        }
      }
    }
    
    if (entry) {
      entry._reloadAvailableCalls = true;
    }
  },
  
  /**************************************************************/
  
  _updateFromCache: function(aIndex) {
    //dump("_updateFromCache\n");
    var entry = this._accountsArray[aIndex];
    var valid = false;
    
    if (entry._en &&
        this._utils.existCache(entry._id)) {
      var contents = this._utils.getCacheContents(entry._id);
      
      if (contents) {
        var err = (contents.hasProperty("error")) ? contents.getPropertyAsString("error") : "";
        // convert NL due to moz-bug #169521
        entry._error = err.replace(/\\n/g, "\n");
        
        entry._phone =
          (contents.hasProperty("phone")) ? contents.getPropertyAsString("phone") : "??";
        
        entry._credit =
          (contents.hasProperty("credit")) ? contents.getPropertyAsString("credit") : "";
        
        entry._messages =
          (contents.hasProperty("messages")) ? contents.getPropertyAsInt("messages") : 0;
        
        entry._dateCallsLoaded =
          (contents.hasProperty("dateCallsLoaded")) ? contents.getPropertyAsInt("dateCallsLoaded") : 0;
        
        if (entry._error.length || entry._credit.length)
          valid = true;
      }
    }
    if (!valid) {
      this._accountsArray[aIndex]._error = "Keine Daten vorhanden.";
      this._accountsArray[aIndex]._credit = "";
    }
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "update::" + entry._id);
  },
  
  /**************************************************************/
  
  _anyForcedUpdates: function() {
    if (this._accountsArray) {
      var i;
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._en &&
            this._accountsArray[i]._forceupdate)
          return i;
      }
    }
    return -1;
  },
  
  /**************************************************************/
  
  _anyQuickLinkUpdates: function() {
    if (this._accountsArray) {
      var i;
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._en &&
            this._accountsArray[i]._quickLinkPending)
          return i;
      }
    }
    return -1;
  },
  
  /**************************************************************/
  
  _needsUpdate: function(aIndex) {
    var entry;
    try {
      entry = this._accountsArray[aIndex];
    }
    catch (e) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "_needsUpdate",
                                "Invalid internal index " + aIndex, false);
      return "";
    }
    
    if (entry._en) {
      if (entry._quickLinkPending)
        return "link";
      
      var now = (new Date()).valueOf();
      if ((entry._dateCallsLoaded + (TO_CALLUPD * this._timeHour) < now) ||
          entry._reloadAvailableCalls)
        return "calls";
      
      if (entry._forceupdate ||
          !this._utils.existCache(entry._id) ||
          this._utils.getCacheValidity(entry._id) == 0)
        return "credit";
    }
    
    return "";
  },
  
  /**************************************************************/
  
  _checkAccounts: function() {
    //dump("_checkAccounts\n");
    try {
      this._checkTimer.cancel();
    }
    catch(e) { }
    
    this._checkPending = false;
    
    if (!this._accountsArray)
      return;
    
    if (this._webAccess.activeSession) {
      this._checkPending = true;
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssSipgateDE._checkAccounts",
                                "Another access already active!",
                                false);
      return;
    }
    
    var type = "link";
    var index = this._anyQuickLinkUpdates();
    
    if (index < 0) {
      index = this._anyForcedUpdates();
      if (index >= 0)
        type = this._needsUpdate(index);
    }
    
    if (index < 0) {
      //dump("no forced updates\n");
      // no forced updates available
      var now = (new Date()).valueOf();
      if ((now - this._lastAccess) > (TO_MIN_PAUSE * this._timeMinute) &&
          now > this._suspendUntil) {
        // ensure a minimum break of TO_MIN_PAUSE minutes between accesses
        //dump("check for scheduled updates\n");
        var i;
        for (i = 0; i < this._accountsArray.length; i++) {
          type = this._needsUpdate(i);
          if (type) {
            //dump("type = " + type + "\n");
            index = i;
            break;
          }
        }
      }
    }
    
    if (index >= 0 && !this._utils.isWebPageOpen(this._blockingUrlRegExp)) {
      this._accessAccount = index;
      this._startWebPath(type);
    }
    else
      this._rescheduleCheck();
  },
  
  /**************************************************************/
  
  _rescheduleCheck: function() {
    //dump("_rescheduleCheck\n");
    if (!this._accountsArray)
      return;
    
    var timeout = 0;
    var blocked = this._utils.isWebPageOpen(this._blockingUrlRegExp);
    
    if (this._checkPending ||
        this._anyQuickLinkUpdates() >= 0 ||
        this._anyForcedUpdates() >= 0) {
      //dump("  immediate check\n");
      if (!blocked) {
        this._checkAccounts();
        return;
      }
    }
    else {
      //dump("  scheduled check\n");
      var now = (new Date()).valueOf();
      var mintimeout = Math.max((TO_MIN_PAUSE * this._timeMinute) - (now - this._lastAccess),
                                this._suspendUntil - now);
      if (mintimeout < 0)
        mintimeout = 0;
      
      timeout = 24 * this._timeHour;              // max. timeout
      for (i = 0; i < this._accountsArray.length; i++) {
        var entry = this._accountsArray[i];
        if (entry._en) {
          if (!this._utils.existCache(entry._id)) {
            //dump("  no cache found for " + entry._id + "\n");
            timeout = 0;
            break;
          }
          else {
            var valid = this._utils.getCacheValidity(entry._id);
            //dump("  cache for " + entry._id + " valid " + valid + "ms\n");
            if (valid < timeout)
              timeout = valid;
          }
        }
      }
      
      if (timeout < mintimeout)
        timeout = mintimeout;
      
      timeout += 1000;               // add 1 second tolerance
    }
    
    if (blocked && (timeout < (TO_BLOCKED * this._timeSecond))) {
      timeout = TO_BLOCKED * this._timeSecond;
      /*this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssSipgateDE._rescheduleCheck",
                                "Scheduled check was delayed by open web page!",
                                false);*/
    }
    
    //dump("timeout = " + timeout + "\n");
    this._checkTimer.initWithCallback(this, timeout,
                                      Components.interfaces.nsITimer.TYPE_ONE_SHOT);
  },
  
  /**************************************************************/
  /**************************************************************/
  
  _webPath: {
    "credit": [ "throttle 1000",
                "login",
                "exit/err",
                "getCredit",
                "exit/err",
                "parseCredit",
                "logout",
                "exit" ],
                
    "calls":  [ "throttle 1000",
                "login",
                "exit/err",
                "getCredit",
                "exit/err",
                "parseCredit",
                "callsPrepare",
                "deposits",
                "depositsDone",
                "jump/err finish",
                "calls",
                "callsDone",
                "jump/err finish",
                "callsStore",
                "finish:",
                "callsComplete",
                "logout",
                "exit" ],
                
    "link":   [ "login",
                "jump/err brerr",
                "loginDone",
                "jump/err brerr",
                "browserGoto",
                "exit",
                "brerr:",
                "browserError",
                "exit" ]
  },
  _webPathType: null,
  
  
  _prevFunctionName: "",
  
  webSeqCallback: function(aFunctionName) {
    //dump(">"+aFunctionName+"\n");
    var prefFN = this._prevFunctionName;
    this._prevFunctionName = aFunctionName;
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "cyCheckyesssSipgateDE.webSeqCallback", aFunctionName, false);
    
    switch (aFunctionName) {
      case "login":            return this._webLogin();
      case "loginDone":        return this._webLoginDone();
      case "getCredit":        return this._webLoginDone("getCredit");
      case "parseCredit":      return this._webCreditDone();
      case "callsPrepare":     return this._webCallsPrepare();
      case "deposits":         return this._webDeposits();
      case "depositsDone":     return this._webDepositsDone();
      case "calls":            return this._webCalls();
      case "callsDone":        return this._webCallsDone();
      case "callsStore":       return this._webCallsStore();
      case "callsComplete":    return this._webCallsComplete();
      case "browserGoto":      return this._webBrowserGoto(false);
      case "browserError":     return this._webBrowserGoto(true);
      case "logout":           return this._webLogout();
      case "exit":             return this._webExit();
      case "error":            return this._webError(prefFN);
      default:
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssSipgateDE.webSeqCallback",
                                  "Unknown function name " + aFunctionName,
                                  false);
        return "";
    }
  },
  
  /**************************************************************/
  
  _startWebPath: function(aPathType) {
    //dump("_startWebPath\n");
    if (!aPathType in this._webPath) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssSipgateDE._startWebPath",
                                "Invalid path type " + aPathType,
                                false);
      return;
    }
    
    if (this._webAccess.activeSession) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssSipgateDE._startWebPath",
                                "Session already active!",
                                false);
      return;
    }
    this._webPathType = aPathType;
    
    this._webAccess.startRequestSequence(new StringEnumerator(this._webPath[aPathType]),
                                         this);
  },
  
  /**************************************************************/
  
  _webLogin: function() {
    //dump("_webLogin\n");
    var entry = this._accountsArray[this._accessAccount];
    
    entry._forceupdate = false;
    this._lastAccess = (new Date()).valueOf();
    
    this._obs.notifyObservers(this, this._observerTopic,
                                    "net-busy::" + entry._id);
    
    if (entry._pw == null) {
      var password = this._managerCenter.getPassword(this._providerId,
                                                     entry._id, entry._user,
                                                     !entry._mayStorePwd, true);
      if (password != null) {
        entry._pw = password;
      }
      else {
        this._webStoreError("Passwort nicht gefunden!");
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssSipgateDE._webLogin",
                                  "Could not find password for account " + entry._id,
                                  false);
        return "err";
      }
    }
    
    var check = new RegExp(this._userRegExp);
    if (!entry._user.match(check)) {
      this._webStoreError("Ung\u00FCltiger Benutzername!");
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssSipgateDE._webLogin",
                                "Invalid format of user name " + entry._user,
                                false);
      return "err";
    }
    
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._sipgateLogin, null, null);
    
    this._webAccess.startRequest(url, "POST",
      "https://secure.sipgate.de/user/index.php",
      "uname=" + encodeURIComponent(entry._user) +
      "&passw=" + encodeURIComponent(entry._pw) +
      "&lasturi=%2Fuser%2Findex.php&okey.x=0&okey.y=0",
      null);
    
    return "ok";
  },
  
  /**************************************************************/
  
  _webLoginDone: function(aFetchCredit) {
    //dump("_webLoginDone\n");
    var result = "ok";
    
    var response = this._webAccess.contentsText;
    var matches = response.match(/\bVisitenkarte\b/i);
    
    if (matches) {
      var entry = this._accountsArray[this._accessAccount];
      entry._error = "";
      
      matches = response.match(/\bVisitenkarte\b.*?\bTelefon\b.*?(0\d+)/i);
      if (matches)
        entry._phone = matches[1];
      else
        entry._phone = "??";
      
      matches = response.match(/(\d+)\s+neue\s+Nachricht/i);
      if (matches)
        entry._messages = Number(matches[1]);
      else
        entry._messages = 0;
      
      if (aFetchCredit) {
        var url = Components.classes["@mozilla.org/network/io-service;1"].
                    getService(Components.interfaces.nsIIOService).
                      newURI(this._sipgateCredit, null, null);
        
        this._webAccess.startRequest(url, "GET",
          "https://secure.sipgate.de/user/start.php", "", null);
      }
    }
    else {
      var error;
      var fail = response.match(/Hilfe-Center\s*(\S.*?)\s+Bitte versuchen Sie es noch/i);
      if (fail) {
        var txt = fail[1].replace(/([^\n]{38}.*?)\s+/g, "$1\n ");
        error = "Fehler!! Sipgate sagt:\n\"" + txt + "\"";
      }
      else
        error = "Fehler beim Einloggen!!";
      
      this._webStoreError(error);
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssSipgateDE._webLoginDone: "+error,
                                this._webAccess.contentsHTML, true);
      
      result = "err";
    }
    
    return result;
  },
  
  /**************************************************************/
  
  _webCreditDone: function() {
    //dump("_webCreditDone\n");
    var result = "ok";
    
    var response = this._webAccess.contentsText;
    var matches = response.match(/\bNeuer\s+Kontostand\b.*?\u20AC\s*(?:(?:-\s*)?(\d+)[,.](\d+))/i);
    
    if (matches) {
      var entry = this._accountsArray[this._accessAccount];
      
      var contents = this._utils.getCacheContents(entry._id);
      if (!contents)
        contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                     createInstance(Components.interfaces.cyICheckyesssContainer);
      
      entry._error = "";
      entry._credit = matches[1]+"."+matches[2];
      this._utils.appendCreditRecord(entry._id, entry._credit);
      contents.setPropertyAsString("phone", entry._phone);
      contents.setPropertyAsString("credit", entry._credit);
      contents.setPropertyAsString("messages", entry._messages);
      contents.setPropertyAsString("error", "");
      
      this._utils.writeCache(entry._id,
                             contents,
                             this._interval * this._timeHour,
                             true);
    }
    else {
      this._webStoreError("Fehler!! Guthaben\nnicht gefunden.");
      result = "err";
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssSipgateDE._webCreditDone: Credit not found!",
                                this._webAccess.contentsHTML, true);
    }
    
    return result;
  },
  
  /**************************************************************/
  
  _callsLastTime: 0,         // time of last stored call record
  _callsCheckTime: 0,        // load calls since _callsCheckTime (overlapping!)
  _callsLoadTime: 0,         // load calls since _callsCheckTime (overlapping!)
  _callsOldet: 0,            // time of oldest calls record loaded
  
  _callsFound: null,         // array of loaded records
  
  _callsMissingFrom: 0,      // start of section missing calls
  _callsMissingTo: 0,        // end of section
  
  
  _webCallsPrepare: function() {
    //dump("_webCallsPrepare\n");
    var entry = this._accountsArray[this._accessAccount];
    var now = new Date();
    var date, loadDate;
    
    if (entry._reloadAvailableCalls) {
      // we HAVE to reload all calls we can find
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssSipgateDE._webCallsPrepare",
                                "Reloading all available calls "+
                                "for account '"+entry._id+"'",
                                false);
      entry._reloadAvailableCalls = false;
      this._callsLastTime = 0;
    }
    else {
      // this is the regular case
      this._callsLastTime = this._utils.getLastCallTime(entry._id);
    }
    
    // store calls overlapping 3 days
    this._callsCheckTime = this._callsLastTime - (3 * 24 * this._timeHour);
    if (this._callsCheckTime > 0) {
      // round to midnight
      date = new Date(this._callsCheckTime);
      date = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
    }
    else {
      /*if (entry._dateCallsLoaded > 0) {
        // apparently we already checked before but did not find any calls
        // -> recheck starting 3 days before that date
        // WARNING: This may lead to problems in the following situation:
        //   Someone mistakenly chose a different phone provider, where the
        //   calls update of course fails. Then she/he corrects the provider
        //   in the settings and now the calls will not be downloaded from
        //   Jan 2nd 2006 but from the date determined one line below!!!!!
        date = new Date(entry._dateCallsLoaded - (3 * 24 * this._timeHour));
        date = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
      }
      else {*/
        // start checking at Jan 2nd 2006
        date = new Date(2006, 0, 2, 0, 0, 0, 0);
      /*}*/
    }
    this._callsCheckTime = date.valueOf();
    date = new Date(date.valueOf() - 1);
    this._callsLoadTime = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0).valueOf();
    this._callsOldest = now.valueOf() + (24 * this._timeHour);
    
    this._callsFound = new Array();
    
    this._callsMissingTo = now.valueOf();
    this._callsMissingFrom = (this._callsLastTime > 0) ? (this._callsLastTime + 1):
                                                         now.valueOf();
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "calls-progress::" + entry._id + "::0");
    return "ok";
  },
  
  /*************/
  
  _webDeposits: function() {
    //dump("_webDeposits\n");
    var now = new Date();
    var loadDate = new Date(this._callsLoadTime);
    
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._sipgateDeposits, null, null);
    
    this._webAccess.startRequest(url, "POST",
      "https://secure.sipgate.at/user/start.php",
      "timeperiod=date"+
      "&timeperiod.begin.day=" + loadDate.getDate() +
      "&timeperiod.begin.month=" + (loadDate.getMonth()+1) +
      "&timeperiod.begin.year=" + loadDate.getFullYear() +
      "&timeperiod.end.day=" + now.getDate() +
      "&timeperiod.end.month=" + (now.getMonth()+1) +
      "&timeperiod.end.year=" + now.getFullYear(),
      null);
    
    return "ok";
  },
  
  /*************/
  
  _webDepositsDone: function() {
    //dump("_webDepositsDone\n");
    var entry = this._accountsArray[this._accessAccount];
    var now = new Date();
    var maxFuture = now.valueOf() + this._timeHour;     // limit time to reasonable values
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "calls-progress::" + entry._id + "::50");
    
    var deposits = this._webAccess.findTable("Aufladungen.*Kontostand",
                                             "(?:\\d{1,2}\\.\\s*\\d{1,2}\\.\\s*\\d{2,4}|heute)");
    if (!deposits) {
      var response = this._webAccess.contentsText;
      // check if there was no any deposit
      if (!response.match(/Aufladungen.*keine.*Kontostand/i)) {
        entry._error = "Fehler beim Auslesen der Gutschriften!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssSipgateDE._webDepositsDone: "+entry._error,
                                  this._webAccess.contentsHTML, true);
        return "err";
      }
      // it is ok to not find any deposits
    }
    
    var onedeposit, fields;
    var matches;
    var hour = 0;
    var minute = 0;
    var time = 0;
    var money = 0;
    var type = "";
    var parseError = null;
    
    while (deposits && deposits.hasMoreElements()) {
      onedeposit = deposits.getNext().QueryInterface(Components.interfaces.nsIStringEnumerator);
      fields = new Array();
      while (onedeposit.hasMore())
        fields.push(onedeposit.getNext());
      
      if (fields.length < 2)
        continue;
      //dump(fields.toString()+"\n");
      
      matches = fields[0].match(/(\d{1,2}):(\d{2})/);
      if (!matches)
        parseError = "time";
      else {
        hour = Number(matches[1]);
        minute = Number(matches[2]);
      }
      if (fields[0].match(/\bheute\b/i)) {
        time = (new Date(now.getFullYear(), now.getMonth(), now.getDate(),
                         hour, minute, 0, 0)).valueOf();
      }
      else {
        matches = fields[0].match(/(\d{1,2})[\.\/]\s*(\d{1,2})[\.\/]\s*(?:20)?(\d{2})/);
        if (!matches)
          parseError = "date";
        else {
          if (matches[3].length < 4)
            matches[3] = "20" + matches[3];
          time = (new Date(Number(matches[3]), Number(matches[2])-1, Number(matches[1]),
                           hour, minute, 0, 0)).valueOf();
          if (time > maxFuture)
            parseError = "reasonable time";
        }
      }
      
      if (time >= this._callsCheckTime && !parseError) {
        var callbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                        createInstance(Components.interfaces.cyICheckyesssContainer);
        
        callbag.setPropertyAsInt("time", time);
        
        type = fields[0];
        type = type.replace(/\d{1,2}\.\s*\d{1,2}\.\s*\d{2,4}|heute/, "");
        type = type.replace(/\d{1,2}:\d{2}(?:\s*h)?/, "");
        type = type.replace(/Aufladung:?\s*/i, "");
        type = type.replace(/^\s*(.*?)\s*$/, "$1");
        
        callbag.setPropertyAsString("type", "reload");
        callbag.setPropertyAsString("ctype", type);
        
        matches = fields[1].match(/(\d+)[,.](\d+)/);
        if (!matches)
          parseError = "credit";
        else
          money = matches[1]+"."+matches[2];
        
        callbag.setPropertyAsString("credit", money);
        
        if (!parseError) {
          this._callsFound.push(callbag);
          if (time < this._callsOldest)
            this._callsOldest = time;
        }
      }
      
      if (parseError) {
        entry._error = "Fehler beim Interpretieren der Gutschriften!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssSipgateDE._webDepositsDone",
                                  "Error parsing "+parseError+":\n"+fields.toString(),
                                  false);
        return "err";
      }
    }
    
    return "done";
  },
  
  /*************/
  
  _webCalls: function() {
    //dump("_webCalls\n");
    var now = new Date();
    var loadDate = new Date(this._callsLoadTime);
    
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._sipgateCalls, null, null);
    
    this._webAccess.startRequest(url, "POST", this._sipgateDeposits,
      "timeperiod=date"+
      "&timeperiod.begin.day=" + loadDate.getDate() +
      "&timeperiod.begin.month=" + (loadDate.getMonth()+1) +
      "&timeperiod.begin.year=" + loadDate.getFullYear() +
      "&timeperiod.end.day=" + now.getDate() +
      "&timeperiod.end.month=" + (now.getMonth()+1) +
      "&timeperiod.end.year=" + now.getFullYear(),
      null);
    
    return "ok";
  },
  
  /*************/
  
  _callsTypeMapping: {
    "sms":        "sms",
    "telefonat":  "call",
    "web-fax":    "connect",
    "gutschrift": "refund",
    "dontknow":   "reload"
  },
  
  _callsTypeSign: {
    "sms":      -1,
    "call":     -1,
    "voicebox": -1,
    "connect":  -1,
    "refund":   +1,
    "reload":   +1
  },
  
  
  _webCallsDone: function() {
    //dump("_webCallsDone\n");
    var entry = this._accountsArray[this._accessAccount];
    var now = new Date();
    var maxFuture = now.valueOf() + this._timeHour;     // limit time to reasonable values
    
    var calls = this._webAccess.findTable("kostenpflichtige.*", "Dauer");
    if (!calls) {
      if (this._webAccess.contentsText.match(/\bEs\s+sind\s+keine\s+Ums/)) {
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                  "cyCheckyesssSipgateDE._webCallsDone",
                                  "There are no calls available from " +
                                  (new Date(this._callsLoadTime)).toString() + " to " +
                                  now.toString(), false);
        return "done";
      }
      else {
        entry._error = "Fehler beim Auslesen der Einzelgespr\u00E4che!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssSipgateDE._webCallsDone: "+entry._error,
                                  this._webAccess.contentsHTML, true);
        return "err";
      }
    }
    
    var onecall, fields;
    var matches, hour, minute;
    var time = 0;
    var pnumber;
    var money = 0;
    var sign = -1;
    var type = "";
    var duration = 0;
    var parseError = null;
    var callcount = 0;
    var date1 = 0;
    
    while (calls.hasMoreElements()) {
      onecall = calls.getNext().QueryInterface(Components.interfaces.nsIStringEnumerator);
      fields = new Array();
      while (onecall.hasMore())
        fields.push(onecall.getNext());
      
      if ((fields.length < 5) || (fields[0] == "Datum"))
        continue;
      
      if (fields.length == 6) {
        // remove the '#' column (the phone line) used in sipgate plus accounts
        fields.splice(2, 1);
      }
      //dump(fields.toString()+"\n");
      
      matches = fields[0].match(/^(.*?)\s+(\d{1,2}):(\d{2})\s*h?$/);
      if (!matches)
        parseError = "date & time";
      else {
        hour = Number(matches[2]);
        minute = Number(matches[3]);
        if (matches[1] == "heute")
          time = (new Date(now.getFullYear(), now.getMonth(), now.getDate(),
                           hour, minute, 0, 0)).valueOf();
        else {
          matches = matches[1].match(/(\d{1,2})[\.\/](\d{1,2})[\.\/](?:20)?(\d{2})/);
          if (!matches)
            parseError = "date";
          else
            time = (new Date(Number("20"+matches[3]), Number(matches[2])-1, Number(matches[1]),
                             hour, minute, 0, 0)).valueOf();
          if (time > maxFuture)
            parseError = "reasonable time";
        }
      }
      //dump("date = "+(new Date(time)).toString()+"\n");
      
      if (time >= this._callsCheckTime && !parseError) {
        var callbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                        createInstance(Components.interfaces.cyICheckyesssContainer);
        
        callbag.setPropertyAsInt("time", time);
        
        // change number to international format
        pnumber = fields[1];
        pnumber = pnumber.replace(/&([a-zA-Z0-9]+)|(#\d+);/g, "");
        pnumber = pnumber.replace(/\s+/g, "");
        if (pnumber.match(/\d/)) {
          pnumber = pnumber.replace(/^00/, "+");            // intl number
          pnumber = pnumber.replace(/^0/, "+49");           // local number
        }
        else {
          pnumber = "";
        }
        callbag.setPropertyAsString("number", pnumber);
        
        matches = fields[4].match(/(\d+)[,.](\d+)/);
        if (!matches)
          parseError = "fee";
        else
          money = matches[1]+"."+matches[2];
        sign = -1;
        
        matches = fields[3].match(/(\d+)?\s*sms/i);
        if (matches) {
          // take care of regular SMS
          type = "sms";
          if (matches[1])
            fields[2] = matches[1] + " " + fields[2];
        }
        else if (fields[2].match(/\bsms\b/i)) {
          // take care of "Voicemail-SMS"
          type = "sms";
          fields[2] = "1 " + fields[2];
        }
        else if (fields[2].match(/web\-?fax/i)) {
          type = "connect";
        }
        else
          type = "call";
        
        callbag.setPropertyAsString("type", type);
        callbag.setPropertyAsString("ctype", fields[2]);
        
        if (sign > 0)
          callbag.setPropertyAsString("credit", money);
        else
          callbag.setPropertyAsString("fee", money);
        
        if (type == "call") {
          matches = fields[3].match(/(?:(?:(\d+)h:?)?(\d+)m:?)?(\d+)s/);
          if (!matches)
            parseError = "duration";
          else {
            if (typeof(matches[1]) == "undefined")
              matches[1] = "0";
            if (typeof(matches[2]) == "undefined")
              matches[2] = "0";
            duration = Number(matches[1])*3600 +
                       Number(matches[2])*60 +
                       Number(matches[3]);
          }
          callbag.setPropertyAsInt("duration", duration);
        }
        
        if (!parseError) {
          this._callsFound.push(callbag);
          if (time < this._callsOldest)
            this._callsOldest = time;
          if (time < this._callsMissingTo)
            this._callsMissingTo = time - 1;
          if (!callcount)               // store the first date for debugging
            date1 = (new Date(time)).toString();
          callcount++;
        }
      }
      
      if (parseError) {
        entry._error = "Fehler beim Interpretieren der Einzelgespr\u00E4che!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssSipgateDE._webCallsDone",
                                  "Error parsing "+parseError+":\n"+fields.toString(),
                                  false);
        return "err";
      }
    }
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "cyCheckyesssSipgateDE._webCallsDone",
                              "Loaded "+callcount+" call records ranging from " +
                              (new Date(date1)).toString() + " to " +
                              (new Date(time)).toString(), false);
    
    return "ok";
  },
  
  /*************/
  
  _webCallsStore: function() {
    //dump("_webCallsStore\n");
    var now = (new Date()).valueOf();
    var entry = this._accountsArray[this._accessAccount];
    
    if (this._callsFound.length > 0) {
      this._utils.storeCallsRecords(entry._id,
                                    new ArrayEnumerator(this._callsFound),
                                    this._callsOldest, now + this._timeHour,
                                    null);
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesssSipgateDE._webCallsStore",
                                "Storing " + this._callsFound.length + " calls",
                                false);
      
      if (this._callsMissingFrom < this._callsMissingTo) {
        this._utils.mergeCallsMissing(entry._id,
                                      this._callsMissingFrom,
                                      this._callsMissingTo);
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                  "cyCheckyesssSipgateDE._webCallsStore",
                                  "Missing calls from "+
                                  (new Date(this._callsMissingFrom)).toString() + " to " +
                                  (new Date(this._callsMissingTo)).toString(), false);
      }
    }
    
    entry._dateCallsLoaded = now;
    var contents = this._utils.getCacheContents(entry._id);
    if (contents) {
      contents.setPropertyAsInt("dateCallsLoaded", entry._dateCallsLoaded);
      this._utils.writeCache(entry._id,              // key
                             contents,
                             this._utils.getCacheValidity(entry._id),
                             false);                 // do not reset age
    }
    
    // we are done -> release some memory
    this._utils.releaseCallsRecords();
    
    return "ok";
  },
  
  /*************/
  
  _webCallsComplete: function() {
    var entry = this._accountsArray[this._accessAccount];
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "calls-progress::" + entry._id + "::100");
    
    if (entry._error) {
      this._webStoreError(entry._error);
    }
    
    return "ok";
  },
  
  /**************************************************************/
  
  _webLogout: function() {
    //dump("_webLogout\n");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._sipgateLogout, null, null);
    this._webAccess.startRequest(url, "GET",
                                 "https://secure.sipgate.de/user/start.php", "",
                                 null);
    
    return "ok";
  },
  
  /**************************************************************/
  
  _webBrowserGoto: function(aError) {  
    //dump("_webBrowserGoto\n");
    var entry = this._accountsArray[this._accessAccount];
    entry._quickLinkPending = false;
    
    if (entry._quickLinkWindow) {
      if (!aError) {
        // display the remote html page
      entry._quickLinkWindow.location = entry._quickLinkTarget;
      }
      else {
        // replace the announcement by an error message
        var errorText = (entry._error) ? entry._error.replace(/\n/, " ") :
                                         "Unbekannter Fehler";
        var errorPage = "<h2><img src='chrome://checkyesss/skin/logo.png'/>"+
              "&nbsp;&nbsp; " + errorText + "</h2></body></html>";
        var body = entry._quickLinkWindow.document.getElementById("qlbody");
        if (body)
          body.innerHTML = errorPage;
      }
      entry._quickLinkWindow.focus();
      entry._quickLinkWindow = null;
      entry._quickLinkTarget = "";
    }
  },
  
  /**************************************************************/
  
  _webStoreError: function(aErrMsg, aFunctionName, aSerious) {
    aFunctionName = (typeof(aFunctionName) != "undefined") ? aFunctionName : "unknown";
    aSerious = (typeof(aSerious) != "undefined") ? aSerious : true;
  
    var entry = this._accountsArray[this._accessAccount];
    
    // A "serious" error is also written into the account's cache causing the
    // icon to turn gray (or similar). Non-serious error are just recorded in
    // the error log.
    if (aSerious) {
    entry._error = aErrMsg;
    
    var contents = this._utils.getCacheContents(entry._id);
    if (!contents)
      contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                   createInstance(Components.interfaces.cyICheckyesssContainer);
    // convert NL due to moz-bug #169521
    contents.setPropertyAsString("error", entry._error.replace(/\n/g, "\\n"));
    
    this._utils.writeCache(entry._id,              // key
                           contents,
                           4 * this._timeHour,     // expires in 4 hours
                           false);                 // do not reset age
    }
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                              "cyCheckyesssSipgateDE._webStoreError",
                              "Account: " + entry._id + "\nCause: " + aErrMsg +
                              "\nFunction: " + aFunctionName + "\nSerious: " + aSerious,
                              false);
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "update::" + entry._id);
  },
  
  /**************************************************************/
  
  _webError: function(aPrefFunctionName) {
    var error = "Unbekannter Fehler";
    var serious = true;
    
    if (this._webAccess.statusText)
      error = this._webAccess.statusText;
    else if (this._webAccess.status > 200)
      error = "Error " + this._webAccess.status;
    
    // Work around a bug (at least that's what it looks right now) in
    // Firefox 3.0b5 (as included in Ubuntu 8.04) where a 302 redirection from
    // a 'https' URL to a 'http' URL results in a security warning while
    // nsIXMLHttpRequest returns a 302 error.
    if (aPrefFunctionName.match(/logout/i))
      serious = false;
    
    this._webStoreError(error, aPrefFunctionName, serious);
    
    return "exit";
  },
  
  /**************************************************************/
  
  _webExit: function() {
    //dump("_webExit\n");
    this._lastAccess = (new Date()).valueOf();
    
    var entry = this._accountsArray[this._accessAccount];
    this._obs.notifyObservers(this, this._observerTopic,
                                    "net-idle::" + entry._id);
    this._obs.notifyObservers(this, this._observerTopic,
                              "update::" + entry._id); //????
    
    if (this._webPathType == "calls") {
      if (entry._error) {
        this._obs.notifyObservers(this, this._observerTopic,
                                  "calls-error::" + entry._id + "::" + entry._error);
      }
      else {
      this._obs.notifyObservers(this, this._observerTopic,
                                "calls::" + entry._id);
    }
    }
    
    if (this._webPathType == "link") {
      entry._quickLinkPending = false;
      entry._quickLinkWindow = null;
      entry._quickLinkTarget = "";
    }
    
    this._rescheduleCheck();
    return "ok";
  },
  
  /**************************************************************/
  
  abortAccess: function() {
    try {
      this._checkTimer.cancel();
    } catch(e) { }
    
    if (this._webAccess.activeSession) {
      //dump("abortAccess\n");
      this._webAccess.abortRequestSequence(false);
      this._lastAccess = (new Date()).valueOf();
      
      var entry = this._accountsArray[this._accessAccount];
      this._obs.notifyObservers(this, this._observerTopic,
                                      "net-idle::" + entry._id);
    }
  },
  
  /**************************************************************/
  
  ///////////////////////////
  // nsITimerCallback
  notify: function(timer) {
    //dump("notify\n");
    if (timer == this._checkTimer) {
      this._checkAccounts();
    }
  },
  
  /**************************************************************/
  
  observe: function(aSubject, aTopic, aData) {
    //do nothing until the first load
    if (this._firstTime)
      return;
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              aTopic, aData, false);
    
    switch (aTopic) {
      case "xpcom-shutdown":
        this._stop();
        break;
      
      default:
        break;
    }
  },
  
  /**************************************************************/
  
  _quickLinkUrls: [
    "http://www.sipgate.de/",
    "https://secure.sipgate.de/user/start.php",
    "https://secure.sipgate.de/user/voicemail.php",
    "https://secure.sipgate.de/user/sms/index.php" ],
  
  jumpQuickLink: function(aId, aIndex, aTab) {
    var win;
    if (aIndex < 1) {
      // no need to log in
      win = this._utils.openUrlInBrowser(this._quickLinkUrls[aIndex], aTab);
      if (win)
        win.focus();
    }
    else if (!this._utils.isThunderbird) {
      // we have to log in before showing the page!
      var entry = null;
      if (this._accountsArray) {
        for (i = 0; i < this._accountsArray.length; i++) {
          if (this._accountsArray[i]._id == aId) {
            if (this._accountsArray[i]._en)
              entry = this._accountsArray[i];
            break;
          }
        }
      }
      
      if (entry) {
        var ps;
        win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
        if (win) {
          ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                     getService(Components.interfaces.nsIPromptService);    
          ps.alert(null, "Achtung",
                   "Die Seite kann nicht aufgerufen werden, da\n"+
                   "bereits ein sipgate-Konto angezeigt wird!\n\n"+
                   "Bitte zuerst ausloggen und dann Aktion wiederholen!");
          
          // focus the mentioned window
          //win.focus();
          
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesssSipgateDE.jumpQuickLink",
                                    "Conflicting browser window detected, action aborted",
                                    false);
        }
        else if (this._anyQuickLinkUpdates() >= 0) {
          ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                     getService(Components.interfaces.nsIPromptService);
          ps.alert(null, "Achtung",
                   "Die Seite kann nicht aufgerufen werden, da\n" +
                   "gerade ein anderer Quick-Link geladen wird!");
          
          // TODO: we could display the window of the active quick link...
          
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesssSipgateDE.jumpQuickLink",
                                    "Another quick link request is already active!",
                                    false);
        }
        else {
          var names = this._managerCenter.getQuickLinkNames(this._providerId,
                                                            "auto");
          var name = null;
          i = aIndex;
          while (names && names.hasMore()) {
            name = names.getNext();
            if (i == 0)
              break;
            i--;
          }
          names = null;
          if (name) {
          var jumpPage = "<html><head>"+
              "<title>" + name + " &ouml;ffnen...</title><head>\n"+
            "<body><h2><img src='chrome://checkyesss/skin/logo.png'/>"+
              "&nbsp;&nbsp;Weiterleitung zu " + name +
            "...</h2></body></html>";
          
          win = this._utils.openUrlInBrowser("", aTab);
          try {
            // does does not work with Firefox 3.0b3
            win.document.open("text/html");
            win.document.write(jumpPage);
            win.document.close();
            win.focus();
          }
          catch (e) { 
            dump(e.toString()+"\n");
          }
          
          entry._quickLinkPending = true;
          entry._quickLinkWindow = win;
          entry._quickLinkTarget = this._quickLinkUrls[aIndex];
          this._checkPending = true;
          
          // restart timer with a short timeout
          this._checkTimer.cancel();
          this._checkTimer.initWithCallback(this, 100, 
                                            Components.interfaces.nsITimer.TYPE_ONE_SHOT);
        }
          else {
            ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                       getService(Components.interfaces.nsIPromptService);
            ps.alert(null, "Achtung",
                     "Es ist ein interner Fehler aufgetreten!");
            
            this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                      "cyCheckyesssSipgateDE.jumpQuickLink",
                                      "Could not find name for quick-link #"+
                                      aIndex+"!", false);
          }
        }
      }
    }
  },
  
  /**************************************************************/
  /**************************************************************/
  
  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.cyICheckyesssManager) &&
        !iid.equals(Components.interfaces.cyICheckyesssWebCallback) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/
var gModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },
  
  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

var gFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new CheckyesssSipgateManager()).QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) { return gModule; }
