/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2006
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{a1bf5bd0-71ba-11db-9fe1-0800200c9a66}");
const CLASS_NAME = "CheckYesss ZipWriter JS Component";
const CONTRACT_ID = "@mozdev.org/checkyesss/zip_writer;1";

/***************************************************************************/

/**
 * The CheckYesss Settings Service.
 */
function CheckyesssZipWriterClass() {
  // allow unwrapping JS component - needed to call setJava()!!!
  this.wrappedJSObject = this;
  this._start();
}


CheckyesssZipWriterClass.prototype = {
  
  _fos: null,
  _zos: null,
  _entryOpen: false,
  
  /**********************************************************************/
  
  _start: function() {
    null;
  },
  
  /**********************************************************************/
  
  _stop: function() {
    // TODO: need to look at how to shut down properly!!!!
    this._fos = null;
    this._zos = null;
    this._java = null;
  },
  
  /**********************************************************************/
  
  setJava: function(aJava) {
    this._java = aJava;
  },
  
  /**********************************************************************/
  
  get error() {
    return this._error;
  },
  
  /**********************************************************************/
  
  createZipFile: function(aZipFileName) {
    try {
      if (typeof(this._java) != "object") {
        this._error = "Fehlende Java-Unterst\u00FCtzung";
        return false;
      }
    }
    catch (e) {
      this._error = "Fehlende Java-Unterst\u00FCtzung"+e;
      return false;
    }
    
    try {
      this._fos = new this._java.io.FileOutputStream(aZipFileName);
    }
    catch (e) {
      this._error = "Fehler beim Anlegen der Datei '"+aZipFileName+"'";
      return false;
    }
    this._zos = new this._java.util.zip.ZipOutputStream(this._fos); 
    this._error = "";
    this._entryOpen = false;
    return true;
  },
  
  /**********************************************************************/
  
  addFile: function(aFileName) {
    if (this._zos) {
      if (this._entryOpen) {
        this._zos.closeEntry();
      }
      var entry = new this._java.util.zip.ZipEntry(aFileName);
      this._zos.putNextEntry(entry);
      this._entryOpen = true;
    }
  },
  
  /**********************************************************************/
  
  addContents: function(aContents, aEncoding) {
    if (this._entryOpen) {
      var bytes = (new this._java.lang.String(aContents)).getBytes(aEncoding);
      this._zos.write(bytes, 0, bytes.length);
    }
  },
  
  /**********************************************************************/
  
  addXMLContents: function(aDOMContents, aEncoding) {
    if (!aEncoding)
      aEncoding = "UTF-8";
    
    if (this._entryOpen) {
      var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
                         createInstance(Components.interfaces.nsIDOMSerializer);
      var string;
      string = serializer.serializeToString(aDOMContents, "text/xml");
      
      // send the string in junks of 1KB
      var bufsize = 1024;
      var index;
      var total;
      var bytes;
      
      // check if we have to add the XML header
      if (string.substr(0, 5) != "<?xml") {
        var header = "<?xml version='1.0' encoding='"+aEncoding+"'?>\n";
        bytes = (new this._java.lang.String(header)).getBytes(aEncoding);
        this._zos.write(bytes, 0, bytes.length);
      }
      
      index = 0;
      total = string.length;
      while (index < total) {
        bytes = (new this._java.lang.String(string.substr(index, bufsize))).getBytes(aEncoding);
        this._zos.write(bytes, 0, bytes.length);
        index += bufsize;
      }
    }
  },
  
  /**********************************************************************/
  
  closeZipFile: function() {
    if (this._zos) {
      if (this._entryOpen) {
        this._zos.closeEntry();
      }
      this._entryOpen = false;
      this._zos.finish();
      this._zos.close();
      this._zos = null;
      this._fos.close();
      this._fos = null;
    }
  },
  
  /**********************************************************************/
  /*          X P C O M - S P E C I F I C   M E T H O D S               */
  /**********************************************************************/

  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.cyICheckyesssZipWriter) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/

var gModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID,
                                     CLASS_NAME,
                                     CONTRACT_ID,
                                     aFileSpec,
                                     aLocation,
                                     aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID) {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) {
    return true;
  }
};

var gFactory = {
  
  createInstance: function (aOuter, aIID) {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (!aIID.equals(Components.interfaces.cyICheckyesssZipWriter) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    
    return ((new CheckyesssZipWriterClass()).QueryInterface(aIID));
  }
};

function NSGetModule(aCompMgr, aFileSpec) {
  return gModule;
}

