// global options dialog window
var gOpWindow = null;


/**************************************/

/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function ArrayEnumerator(array) {
    this.array = array;
    this.index = 0;
}
ArrayEnumerator.prototype = {
    hasMoreElements: function() { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};

/**************************************/

function doLoad() {
  gOpWindow = new OpWindow();
  gOpWindow.start();
  return true;
}


function doUnload() {
  gOpWindow.stop();
  gOpWindow = null;
}


function OpWindow() {
  return this;
}


OpWindow.prototype = {
  
  _managerCenter: null,
  _utils: null,
  
  _globalSettings: null,
  _accountSettings: null,
  _deletedAccounts: null,
  _iconOrder: null,           // numerical order of the icons
  _idOrder: null,             // alphabetical order of the IDs
  
  _iconsEnabled: 0,           // number of enabled/shown icons
  
  _selectedAccount: null,     // container of currently selected account in accounts panel
  _selectedPanel: null,       // currently selected statusbarpanel
  _validUserRegExp: null,
  
  _settingsChanged: false,
  _blinkTimer: null,
  
  
  start: function() {
    this._utils = Components.classes["@mozdev.org/checkyesss/util_service;1"].
                getService(Components.interfaces.cyICheckyesssUtil);
    
    this._managerCenter = Components.classes["@mozdev.org/checkyesss/manager_center;1"].
                            getService(Components.interfaces.cyICheckyesssCenter);
    
    this._settings = Components.classes["@mozdev.org/checkyesss/settings_service;1"].
                       getService(Components.interfaces.cyICheckyesssSettings);
    
    this._initDialog();
  },
  
  
  stop: function() {
    //stop timers
    if (this._blinkTimer)
      clearTimeout(this._blinkTimer);
    
    //destroy variables
    this._utils = null;
    this._globalSettings = null;
    this._accountSettings = null;
    this._deletedAccounts = null;
    this._iconOrder = null;
    this._idOrder = null;
    this._selectedAccount = null;
    this._selectedPanel = null;
    this._blinkTimer = null;
    this._validUserRegExp = null;
  },
  
  
  // OK button pressed
  doOK: function() {
    if (this._settingsChanged) {
      var i, ac;
      var oldId, newId, userId;
      var manager, account, provider;
      var managerIds;
      var managerId;
      
      // collect all managers involved
      managerIds = new Object();
      for (ac = 0; ac < this._deletedAccounts.length; ac++) {
        // the _original_ provider is relevant here, not the current one
        // which was possibly changed by the user before deleting the account!
        oldId = this._deletedAccounts[ac].getPropertyAsString("original-id");
        account = this._settings.getAccountSettings(oldId);
        if (account) {
          // this account really existed before
          provider = account.getPropertyAsString("provider");
          managerIds[provider] = 1;
        }
      }
      for (ac = 0; ac < this._accountSettings.length; ac++) {
        // same as above concerning the _original_ provider
        oldId = this._accountSettings[ac].getPropertyAsString("original-id");
        account = this._settings.getAccountSettings(oldId);
        if (account) {
          // this account really existed before
          provider = account.getPropertyAsString("provider");
          managerIds[provider] = 1;
        }
      }
      // clear all accounts of these managers
      for (managerId in managerIds) {
        manager = this._managerCenter.getManager(managerId);
        if (manager)
          manager.clearAccounts();
        this._managerCenter.releaseManager(managerId);
      }
      
      // Clean up data of deleted accounts
      for (ac = 0; ac < this._deletedAccounts.length; ac++) {
        oldId = this._deletedAccounts[ac].getPropertyAsString("original-id");
        account = this._settings.getAccountSettings(oldId);
        if (account) {
          // this account really existed before
          this._utils.removeCreditLog(oldId);
          this._utils.removeCallsRecords(oldId);
          this._utils.removeCache(oldId);
          // if the original settings asked to store the password then remove it
          if (account.getPropertyAsString("pw-handling") == "store")
            this._managerCenter.removePassword(account.getPropertyAsString("provider"),
                                               oldId,
                                               account.getPropertyAsString("user"));
        }
      }
      
      // Next: Take care about renamed accounts
      for (ac = 0; ac < this._accountSettings.length; ac++) {
        oldId = this._accountSettings[ac].getPropertyAsString("original-id");
        newId = this._accountSettings[ac].getPropertyAsString("id");
        account = this._settings.getAccountSettings(oldId);
        
        if (account) {
          // this account really existed before
          if (oldId != newId) {
            // .. and changed its ID
            this._utils.renameCreditKey(oldId, newId);
            this._utils.renameCallsRecords(oldId, newId);
            if (this._utils.existCache(oldId))
              this._utils.renameCacheKey(oldId, newId, true);
          }
          
          // if the original settings asked to store the password then remove it
          // even if it gets re-added later
          if (account.getPropertyAsString("pw-handling") == "store")
            this._managerCenter.removePassword(account.getPropertyAsString("provider"),
                                               oldId,
                                               account.getPropertyAsString("user"));
        }
      }
      
      // need to extract the passwords as they will be removed
      // by this._settings.writeNewSettings!!
      var pw = new Array();
      for (ac = 0; ac < this._accountSettings.length; ac++) {
        if (this._accountSettings[ac].getPropertyAsString("pw-handling") == "store")
          pw[ac] = this._accountSettings[ac].getPropertyAsString("_password");
        else
          pw[ac] = null;
        this._accountSettings[ac].deleteProperty("_password");
      }
      
      // special treatment of certain settings
      var newint = document.getElementById("options-interval").getAttribute("value");
      this._globalSettings.setPropertyAsInt("update-interval", Number(newint));
      
      // write settings to file
      this._settings.writeNewSettings(this._globalSettings,
                                      new ArrayEnumerator(this._accountSettings));
      
      // (re-)add accounts to account managers & password manager
      for (ac = 0; ac < this._accountSettings.length; ac++) {
        account = this._accountSettings[ac];
        provider = account.getPropertyAsString("provider");
        newId = account.getPropertyAsString("id");
        userId = account.getPropertyAsString("user");
        if (pw[ac] != null)
          this._managerCenter.storePassword(provider, newId, userId, pw[ac]);
        manager = this._managerCenter.getManager(provider);
        manager.addAccount(newId,
                           userId,
                           pw[ac],
                           account.getPropertyAsString("rateplan"),
                           account.getPropertyAsString("period-reset"),
                           account.getPropertyAsBool("enabled"),
                           (account.getPropertyAsString("pw-handling") == "store"));
      }
    }
    //else
      //dump("doOk: NO changes!\n");
    
    window.close();
    return true;
  },
  
  
  // Cancel button pressed
  doCancel: function() {
    window.close();
    return true;
  },
  
  
  /**
   * Initialize the dialog with global and account settings
   */
  
  _initDialog: function() {
    var value;
    var isProviderOk, account, provider;
    var id, ids, rateplanIndex;
    var rateplan, rateplanIds, index;
    
    // check if the "Add Account" should be displayed automatically
    var autoAdd = (window.arguments && window.arguments[0] == "add");
    
    /* switch page */
    
    var pageSelector = document.getElementById("options-categories");
    value = (autoAdd) ? 1 : pageSelector.getAttribute("mySelectedIndex");
    pageSelector.selectedIndex = Number(value);
    
    var title = pageSelector.selectedItem.getAttribute("label");
    document.getElementById("options-deck").selectedIndex = pageSelector.selectedIndex;
    document.getElementById("options-header").setAttribute("title", title);
    
    /** load the current settings **/
    
    this._globalSettings = this._settings.getGlobalSettings().clone();
    
    this._accountSettings = new Array();
    this._deletedAccounts = new Array();
    
    ids = this._settings.getAccountIds();
    while (ids.hasMore()) {
      id = ids.getNext();
      isProviderOk = false;
      rateplanIndex = -1;
      account = this._settings.getAccountSettings(id).clone();
      // copy id to original-id to detect name changes!
      account.setPropertyAsString("original-id", id);
      if (account.hasProperty("provider")) {
        provider = account.getPropertyAsString("provider");
        isProviderOk = this._managerCenter.isValidProvider(provider);
        if (isProviderOk) {
          rateplan = account.getPropertyAsString("rateplan");
          rateplanIds = this._managerCenter.getRatePlanList(provider);
          index = 0;
          while (rateplanIds.hasMore()) {
            if (rateplanIds.getNext() == rateplan) {
              rateplanIndex = index;
              break;
            }
            index++;
          }
        }
      }
      if (!isProviderOk || rateplanIndex < 0) {
        account.setPropertyAsBool("enabled", false);
        account.setPropertyAsString("_password", "");
        if (!isProviderOk) {
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "options._initDialog",
                                    "Invalid provider '" + account.getPropertyAsString("provider") +
                                    "', setting to default provider",
                                    false);
          
          // default to first manager
          provider = this._managerCenter.getProviderList.getNext();
          account.setPropertyAsString("provider", provider);
        }
        if (rateplanIndex < 0) {
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "options._initDialog",
                                    "Invalid rate plan '" + account.getPropertyAsString("rateplan") +
                                    "', setting to default rate plan",
                                    false);
          rateplanIndex = 0;         // default to first rate plan
          rateplan = this._managerCenter.getRatePlanList(provider).getNext();
          account.setPropertyAsString("rateplan", rateplan);
        }
        this._settingsChanged = true;
      }
      else {
        var pw = null;
        if (account.getPropertyAsString("pw-handling") == "store")
          pw = this._managerCenter.getPassword(provider, id,
                         account.getPropertyAsString("user"), false, false);
        if (pw == null)
          pw = "";
        account.setPropertyAsString("_password", pw);
      }
      
      /* additional properties which are not stored on disk */
      account.setPropertyAsInt("_rateplannum", rateplanIndex);
      
      this._accountSettings.push(account);
    }
    this._sortAndCompress();
    
    /** apply settings to dialog elements **/
    
    /* general settings */
    
    value = this._globalSettings.getPropertyAsInt("update-interval");
    document.getElementById("options-interval-slider").
      setAttribute("curpos", value - 4);
    document.getElementById("options-interval").
      setAttribute("value", value);
    document.getElementById("options-interval-slider").
      addEventListener("DOMAttrModified", gOpWindow._updateInterval, false);
    
    document.getElementById("options-nohide").checked =
      !this._globalSettings.getPropertyAsBool("autohide-all");
    
    document.getElementById("options-divider").checked =
      this._globalSettings.getPropertyAsBool("show-divider");
    
    document.getElementById("options-whatsnew").checked =
      this._globalSettings.getPropertyAsBool("show-whatsnew");
    
    /* mouse settings */
    
    document.getElementById("options-linksmenu").checked =
      this._globalSettings.getPropertyAsBool("linksmenu-tab");
    
    /* debug settings */
    
    var logfile = this._utils.logFile.clone();
    document.getElementById("options-logsize").value = 
      this._utils.convertFileSize(logfile.fileSize);
    document.getElementById("options-logname").value = logfile.path;
    document.getElementById("options-loglevel").selectedIndex = 
      this._globalSettings.getPropertyAsInt("loglevel");
    
    /* current account */
    
    this._updateManagerList();
    this._updateIconList();
    this._updateAccountList();
    this._switchAccount(0);
    this._updatePreview();
    
    if (window.arguments && window.arguments[0] == "add")
      value = 1;
    else
      value = pageSelector.getAttribute("mySelectedIndex");
    
    if (autoAdd)
      // display the "Add Account" dialog
      setTimeout("gOpWindow._addAccount();", 100);
  },
  
  /****************/
  
  updateRateplanTooltip: function(event) {
    var box = document.getElementById("rateplan-tooltip-box");
    while (box.lastChild)
      box.removeChild(box.lastChild);
    
    var item = document.getElementById("account-rateplan").selectedItem;
    
    if (item && item.hasAttribute("tooltiptext")) {
      var text = item.getAttribute("tooltiptext");
      if (text.length > 0) {
        var lines = text.split("\n");
        for (var i = 0; i < lines.length; i++) {
          var descr = document.createElement("description");
          var ltext = document.createTextNode(lines[i]);
          descr.appendChild(ltext);
          box.appendChild(descr);
        }
        return true;
      }
    }
    return false;
  },
  
  /****************/
  
  switchPage: function(event) {
    var title = event.target.selectedItem.getAttribute("label");
    var index = event.target.selectedIndex;
    document.getElementById("options-deck").selectedIndex = index;
    document.getElementById("options-header").setAttribute("title", title);
    event.target.setAttribute("mySelectedIndex", index);
    return true;
  },
  
  /****************/
  
  _sortAndCompress: function() {
    var refOrg = new Array();
    var refNew = new Array();
    var i, j;
    var len = this._accountSettings.length;
    var max;
    
    /* sort icon positions */
    
    var numsort = function(a, b) {
      return a-b;
    };
    
    // get the positions but move disabled accounts to the end!
    max = 0;
    for (i = 0; i < len; i++) {
      refOrg[i] = this._accountSettings[i].getPropertyAsInt("position");
      if (max < refOrg[i])
        max = refOrg[i];
    }
    max++;
    for (i = 0; i < len; i++) {
      if (!this._accountSettings[i].getPropertyAsBool("enabled"))
        refOrg[i] += max;
      refNew[i] = refOrg[i];
    }
    
    refNew.sort(numsort);
    this._iconOrder = new Array();
    
    for (i = 0; i < len; i++) {
      for (j = 0; j < len; j++) {
        if (refOrg[i] == refNew[j])
          break;
      }
      this._accountSettings[i].setPropertyAsInt("position", j);
      this._iconOrder[j] = i;
      refNew[j] = -1;    // invalidate
    }
    
    /* sort IDs */
    
    this._idOrder = new Array();
    refNew = new Array();
    
    for (i = 0; i < len; i++)
      refNew[i] = this._accountSettings[i].getPropertyAsString("id");
    
    refNew.sort();
    
    for (i = 0; i < len; i++) {
      var id = this._accountSettings[i].getPropertyAsString("id");
      for (j = 0; j < len; j++) {
        if (id == refNew[j])
          break;
      }
      this._idOrder[j] = i;
    }
  },
  
  /****************/
  
  _updateManagerList: function() {
    var menulist = document.getElementById("account-provider");
    menulist.selectedIndex = -1;
    while (menulist.removeItemAt(0)) { null; }
    //menulist.removeAllItems();    breaks things
    
    var managerIds = this._managerCenter.getProviderList;
    var managerId, name;
    
    while (managerIds.hasMore()) {
      managerId = managerIds.getNext();
      name = this._managerCenter.getProviderProperties(managerId).
               getPropertyAsString("name");
      menulist.appendItem(name, managerId);
    }
  },
  
  /****************/
  
  _updateRatePlanList: function(aProvider) {
    var menulist = document.getElementById("account-rateplan");
    menulist.selectedIndex = -1;
    while (menulist.removeItemAt(0)) { null; }
    //menulist.removeAllItems();    breaks things
    
    if (aProvider) {
      var ratePlanIds = this._managerCenter.getRatePlanList(aProvider);
      while (ratePlanIds.hasMore()) {
        var id = ratePlanIds.getNext();
        var plan = this._managerCenter.getRatePlanProperties(aProvider, id);
        var item = menulist.appendItem(plan.getPropertyAsString("name"), id);
        /*var descr = plan.getPropertyAsString("description");
        if (descr !== null && descr.length)
          item.setAttribute("tooltiptext", descr);*/
      }
    }
    else {
      menulist.appendItem("", "");
    }
  },
  
  /****************/
  
  _updateIconList: function() {
    var menulist = document.getElementById("account-icon");
    menulist.selectedIndex = -1;
    while (menulist.removeItemAt(0)) { null; }
    //menulist.removeAllItems();    breaks things
    
    var iconNames = this._settings.getIconNames();
    while (iconNames.hasMore()) {
      var iconName = iconNames.getNext();
      var iconProp = this._settings.getIconProperties(iconName);
      var previewUrl = iconProp.getPropertyAsString("preview");
      
      var menuitem = menulist.appendItem(null, iconName);
      var image = document.createElement("image");
      image.setAttribute("src", previewUrl);
      menuitem.appendChild(image);
      menuitem.setAttribute("image", previewUrl);  // needed by Firefox 3.0
      menuitem.setAttribute("src", previewUrl);    // needed by Seamonkey
    }
  },
  
  /****************/
  
  _updateAccountList: function() {
    var menulist = document.getElementById("account-list");
    menulist.selectedIndex = -1;
    while (menulist.removeItemAt(0)) { null; }
    //menulist.removeAllItems();    breaks things
    
    if (this._accountSettings.length) {
      for (var i = 0; i < this._accountSettings.length; i++) {
        var index = this._idOrder[i];
        var id = this._accountSettings[index].getPropertyAsString("id");
        menulist.appendItem(id, ""+i);
      }
      document.getElementById("account-header").removeAttribute("blink");
      this._accountHeaderBlink();
    }
    else {
      menulist.appendItem("", "-1");
      document.getElementById("account-header").setAttribute("blink", "true");
      this._accountHeaderBlink();
    }
  },
  
  /****************/
  
  _accountHeaderBlink: function() {
    var element = document.getElementById("account-header");
    if (!element.getAttribute("blink")) {
      element.removeAttribute("style");
      if (this._blinkTimer)
        clearTimeout(this._blinkTimer);
      this._blinkTimer = null;
    }
    else {
      if (element.getAttribute("style"))
        element.removeAttribute("style");
      else
        element.setAttribute("style", "background-color:red;");
      if (this._blinkTimer)
        clearTimeout(this._blinkTimer);
      this._blinkTimer = setTimeout("gOpWindow._accountHeaderBlink();", 500);
    }
  },
  
  /****************/
  
  _updatePreview: function() {
    var defaults = [
      "imgclass",   "checkyesss-anim-icon",
      "charclass",  "checkyesss-anim-char",
      "labelclass", "statusbarpanel-text",
      "label",      "12,34",
      "imgmood",    "ok",
      "animate",    "false",
      "delay",      "100",
      "imgcnt",     "0" ];
    var innerbox = document.getElementById("icon-inner-box");
    innerbox.setAttribute("visibility", "collapsed");
    while (innerbox.hasChildNodes()) {
      innerbox.removeChild(innerbox.lastChild);
    }
    this._iconsEnabled = 0;
    
    if (this._accountSettings.length) {
      var isFirst = true;
      for (var i = 0; i < this._accountSettings.length; i++) {
        var index = this._iconOrder[i];
        var account = this._accountSettings[index];
        
        if (account.getPropertyAsBool("enabled")) {
          if ((this._iconsEnabled > 0) &&
              this._globalSettings.getPropertyAsBool("show-divider")) {
            var divider = document.createElement("box");
            divider = innerbox.appendChild(divider);
            divider.className = "checkyesss-panel-separator";
          }
          
          var id = account.getPropertyAsString("id");
          var panel = document.createElement("hbox");
          panel.setAttribute("id", "options-preview-" + index);
          for (var d = 0; d < defaults.length; d += 2)
            panel.setAttribute(defaults[d], defaults[d+1]);
          panel.setAttribute("char", account.getPropertyAsString("initial"));
          
          var iconName = account.getPropertyAsString("icon-theme");
          var iconProp = this._settings.getIconProperties(iconName);
          panel.setAttribute("imgtheme", iconName);
          if (iconProp.getPropertyAsBool("animated")) {
            panel.setAttribute("delay", 
                               iconProp.getPropertyAsInt("anim-delay"));
            panel.setAttribute("imgcnt", 
                               iconProp.getPropertyAsInt("anim-count"));
          }
          
          var showlabel = account.getPropertyAsString("show-label");
          if (showlabel == "autoShow")    // no support for autoShow here
            showlabel = "autoShowHide";
          if (iconProp.getPropertyAsBool("hidden"))
            showlabel = "yes";
          panel.setAttribute("showlabel", showlabel);
          panel.setAttribute("tooltiptext", account.getPropertyAsString("id"));
          
          panel = innerbox.appendChild(panel);
          panel.className = "checkyesss-panel";
          
          this._iconsEnabled++;
        }
      }
    }
    
    innerbox.setAttribute("visibility", "visible");
    this._updatePreviewSelect();
  },
  
  /****************/
  
  _updatePreviewSelect: function() {
    this._selectedPanel = null;
    
    if (this._iconsEnabled > 1) {
      document.getElementById("icon-move-left").removeAttribute("disabled");
      document.getElementById("icon-move-right").removeAttribute("disabled");
    }
    else {
      document.getElementById("icon-move-left").disabled = true;
      document.getElementById("icon-move-right").disabled = true;
    }
    
    if (this._iconsEnabled > 0) {
      var innerbox = document.getElementById("icon-inner-box");
      var child = innerbox.firstChild;
      var firstPanel = null;
      var lastPanel = null;
      while (child) {
        if (child.className == "checkyesss-panel") {
          var index = Number(child.id.replace(/^options-preview-/, ""));
          if (this._selectedAccount == this._accountSettings[index]) {
            child.setAttribute("selected", "true");
            this._selectedPanel = child;
            if (!firstPanel)
              document.getElementById("icon-move-left").disabled = true;
          }
          else
            child.removeAttribute("selected");
          
          if (!firstPanel)
            firstPanel = child;
          lastPanel = child;
        }
        child = child.nextSibling;
      }
      if (lastPanel.getAttribute("selected") == "true")
        document.getElementById("icon-move-right").disabled = true;
    }
  },
  
  /****************/
  
  _switchAccount: function(aIndex) {
    if (aIndex >= this._accountSettings.length)
      aIndex = this._accountSettings.length - 1;
    
    var broadcaster = document.getElementById("noaccount-broadcaster");
    
    if (aIndex < 0) {
      this._selectedAccount = null;
      
      broadcaster.setAttribute("disabled", "true");
      
      this._updateRatePlanList(null);
      document.getElementById("account-list").selectedIndex = 0;
      document.getElementById("account-enabled").checked = false;
      document.getElementById("account-provider").selectedIndex = 0;
      document.getElementById("account-rateplan").selectedIndex = 0;
      document.getElementById("account-login").value = "";
      document.getElementById("account-login-label").removeAttribute("style");
      document.getElementById("account-password").value = "";
      document.getElementById("account-pw-handling").selectedIndex = -1;
      
      document.getElementById("account-icon").selectedIndex = 0;
      document.getElementById("account-animate").checked = false;
      document.getElementById("account-initial").value = "";
      document.getElementById("account-autoshow").selectedIndex = -1;
      document.getElementById("account-showage").checked = false;
      
      this._updateMouseActionLists(null, null);
      document.getElementById("options-ldest").selectedIndex = 0;
      document.getElementById("options-mdest").selectedIndex = 0;
      document.getElementById("options-lurl").value = "";
      document.getElementById("options-murl").value = "";
      document.getElementById("options-ltab").checked = false;
      document.getElementById("options-mtab").checked = false;
      
      document.getElementById("tab-period").setAttribute("hidden", "true");
    }
    else {
      var account, provider, index;
      var menulist, menuitem;
      
      broadcaster.removeAttribute("disabled");
      
      document.getElementById("account-list").selectedIndex = aIndex;
      
      index = this._idOrder[aIndex];
      this._selectedAccount = this._accountSettings[index];
      account = this._selectedAccount;
      provider = account.getPropertyAsString("provider");
      
      this._changeRateplan(account.getPropertyAsString("rateplan"));
      
      /* -- Login-Data tab -- */
      this._updateRatePlanList(provider);
      
      menulist = document.getElementById("account-provider");
      index = 0;
      menulist.selectedIndex = 0;
      menuitem = menulist.selectedItem;
      while (menuitem) {
        if (menuitem.value == provider) {
          menulist.selectedIndex = index;
          break;
        }
        menuitem = menuitem.nextSibling;
        index++;
      }
      
      document.getElementById("account-enabled").checked =
        account.getPropertyAsBool("enabled");
      document.getElementById("account-rateplan").selectedIndex =
        account.getPropertyAsInt("_rateplannum");
      document.getElementById("account-login").value =
        account.getPropertyAsString("user");
      document.getElementById("account-password").value =
        account.getPropertyAsString("_password");
      
      if (account.getPropertyAsString("pw-handling") == "store") {
        document.getElementById("account-pw-handling").selectedIndex = 0;
        document.getElementById("account-password").removeAttribute("disabled");
      }
      else {
        document.getElementById("account-pw-handling").selectedIndex = 1;
        document.getElementById("account-password").setAttribute("disabled", "true");
      }
      
      var regexpStr = this._managerCenter.getProviderProperties(provider).
                        getPropertyAsString("validUserRegExp");
      this._validUserRegExp = new RegExp(regexpStr);
      if (!account.getPropertyAsString("user").match(this._validUserRegExp))
        document.getElementById("account-login-label").setAttribute("style", "color: red;");
      else
        document.getElementById("account-login-label").removeAttribute("style");
      
      /* -- Display tab -- */
      /* find the account's icon in the list */
      var iconname = account.getPropertyAsString("icon-theme");
      var iconlist = document.getElementById("account-icon");
      var iconindex = 0;
      iconlist.selectedIndex = 0;
      menuitem = iconlist.selectedItem;
      while (menuitem) {
        if (menuitem.value == iconname) {
          iconlist.selectedIndex = iconindex;
          break;
        }
        iconindex++;
        menuitem = menuitem.nextSibling;
      }
      iconname = iconlist.selectedItem.value;
      account.setPropertyAsString("icon-theme", iconname);
      
      if (this._settings.getIconProperties(iconname).getPropertyAsBool("animated")) {
        document.getElementById("account-animate").removeAttribute("disabled");
        document.getElementById("account-animate").checked =
          account.getPropertyAsBool("animate");
      }
      else {
        document.getElementById("account-animate").setAttribute("disabled", "true");
        document.getElementById("account-animate").checked = false;
        account.setPropertyAsBool("animate", false);
      }
      
      document.getElementById("account-initial").value =
        account.getPropertyAsString("initial");
      
      var acautoshow = document.getElementById("account-autoshow");
      if (this._settings.getIconProperties(iconname).getPropertyAsBool("hidden")) {
        acautoshow.selectedIndex = 0;
        document.getElementById("account-autoshow-no").setAttribute("disabled", "true");
        document.getElementById("account-autoshow-auto").setAttribute("disabled", "true");
      }
      else {
        var autoshow = account.getPropertyAsString("show-label");
        acautoshow.selectedIndex =
          (autoshow == "no") ? 1 :
          (autoshow == "autoShow") ? 2 : 0;
        autoshow = acautoshow.selectedItem.value;
        account.setPropertyAsString("show-label", autoshow);
        document.getElementById("account-autoshow-no").removeAttribute("disabled");
        document.getElementById("account-autoshow-auto").removeAttribute("disabled");
      }
      
      document.getElementById("account-showage").checked =
        account.getPropertyAsBool("show-age");
      
      /* -- Quick-Link tab -- */
      //this._updateMouseActionLists(provider, account.getPropertyAsString("rateplan"));
      
      value = account.getPropertyAsString("lclick-dest");
      menulist = document.getElementById("options-ldest");
      menulist.selectedIndex = 0;
      while (menulist.selectedItem) {
        if (menulist.selectedItem.value == value)
          break;
        else
          menulist.selectedIndex++;
      }
      if (!menulist.selectedItem) {
        menulist.selectedIndex = 0;
        value = menulist.selectedItem.value;
        account.setPropertyAsString("lclick-dest", value);
      }
      document.getElementById("options-lurl").disabled = (value != "user");
      document.getElementById("options-lurl").value =
        account.getPropertyAsString("lclick-url");
      document.getElementById("options-ltab").checked =
        account.getPropertyAsBool("lclick-tab");
      
      value = account.getPropertyAsString("mclick-dest");
      menulist = document.getElementById("options-mdest");
      menulist.selectedIndex = 0;
      while (menulist.selectedItem) {
        if (menulist.selectedItem.value == value)
          break;
        else
          menulist.selectedIndex++;
      }
      if (!menulist.selectedItem) {
        menulist.selectedIndex = 0;
        value = menulist.selectedItem.value;
        account.setPropertyAsString("mclick-dest", value);
      }
      document.getElementById("options-murl").disabled = (value != "user");
      document.getElementById("options-murl").value =
        account.getPropertyAsString("mclick-url");
      document.getElementById("options-mtab").checked =
        account.getPropertyAsBool("mclick-tab");
      
      /* -- Accounting Period -- */
      var preset = account.getPropertyAsString("period-reset");
      var dayofmonth = preset.match(/^\s*(\d+)/);
      if (dayofmonth) {
        document.getElementById("period-reset").selectedIndex = 1;
        document.getElementById("period-mday").removeAttribute("disabled");
        document.getElementById("period-mday").selectedIndex = dayofmonth[1] - 1;
      }
      else {
        document.getElementById("period-reset").selectedIndex =
          (preset == "auto") ? 0 : 2;
        document.getElementById("period-mday").disabled = true;
        document.getElementById("period-mday").selectedIndex = 0;
      }
    }
  },
  
  /****************/
  
  _addAccount: function() {
    var args = { type: "add",
                 title: "Konto hinzuf\u00FCgen",
                 exist: [ ],
                 deleted: [ ],
                 result: "" };
    var i;
    
    if (this._accountSettings) {
      for (i = 0; i < this._accountSettings.length; i++)
        args.exist.push(this._accountSettings[i].getPropertyAsString("id"));
    }
    if (this._deletedAccounts) {
      for (i = 0; i < this._deletedAccounts.length; i++)
        args.deleted.push(this._deletedAccounts[i].getPropertyAsString("id"));
    }
    
    window.openDialog("chrome://checkyesss/content/options-name.xul",
                      "checkyesss-opdlg",
                      "chrome,modal", args);
    
    var newId = args.result;
    if (newId.length > 0) {
      var account = null;
      var provider;
      // check whether user wants to recover a deleted account
      if (this._deletedAccounts) {
        for (i = 0; i < this._deletedAccounts.length; i++) {
          if (newId == this._deletedAccounts[i].getPropertyAsString("id")) {
            account = this._deletedAccounts[i];
            if (i < this._deletedAccounts.length-1)
              this._deletedAccounts[i] = this._deletedAccounts.pop();
            else
              this._deletedAccounts.pop();
          }
        }
      }
      if (!account) {
        // this is a completely new account
        account = this._settings.getAccountDefaults();
        account.setPropertyAsString("id", newId);
        account.setPropertyAsString("original-id", newId);
        account.setPropertyAsString("_password", "");
        
        if (this._accountSettings != null &&
            this._accountSettings.length > 0) {
          // copy the provider from one of the existing accounts
          provider = this._accountSettings[0].getPropertyAsString("provider");
        }
        else
          provider = this._managerCenter.getProviderList.getNext();
        
        account.setPropertyAsString("provider", provider);
        account.setPropertyAsString("rateplan",
          this._managerCenter.getRatePlanList(provider).getNext());
        account.setPropertyAsInt("_rateplannum", 0);
        
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                  "options._addAccount",
                                  "Adding new account '" + newId + "'",
                                  false);
      }
      else {
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                  "options._addAccount",
                                  "Recovering deleted account '" + newId + "'",
                                  false);
      }
      account.setPropertyAsInt("position", this._accountSettings.length);
      account.setPropertyAsBool("enabled", true);
      
      this._accountSettings.push(account);
      
      this._sortAndCompress();
      this._updateAccountList();
      
      for (i = 0; i < this._accountSettings.length; i++) {
        if (this._idOrder[i] == this._accountSettings.length-1) {
          this._switchAccount(i);
          break;
        }
      }
      
      this._updatePreview();
    }
    else {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "options._addAccount",
                                "Cancelled dialog",
                                false);
    }
  },
  
  /****************/
  
  _renameAccount: function(aIndex) {
    var index = this._idOrder[aIndex];
    var account = this._accountSettings[index];
    var oldId = account.getPropertyAsString("id");
    var i;
    
    var args = { type: "rename",
                 title: "Name des Kontos '"+oldId+"' \u00E4ndern",
                 exist: [ ],
                 oldname: oldId,
                 result: "" };
    if (this._accountSettings) {
      for (i = 0; i < this._accountSettings.length; i++) {
        if (i != index)
          args.exist.push(this._accountSettings[i].getPropertyAsString("id"));
      }
    }
    
    window.openDialog("chrome://checkyesss/content/options-name.xul",
                      "checkyesss-opdlg",
                      "chrome,modal", args);
    
    var newId = args.result;
    if (newId.length > 0 && newId != oldId) {
        account.setPropertyAsString("id", newId);
      
      this._sortAndCompress();
      this._updateAccountList();
      
      for (i = 0; i < this._accountSettings.length; i++) {
        if (this._idOrder[i] == index) {
          this._switchAccount(i);
          break;
        }
      }
      
      this._updatePreview();
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "options._renameAccount",
                                "Renamed account '" + oldId + "' to '" + newId + "'",
                                false);
    }
    else {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "options._renameAccount",
                                "Cancelled dialog",
                                false);
    }
  },
  
  /****************/
  
  _deleteAccount: function(aIndex) {
    var index = this._idOrder[aIndex];
    var account = this._accountSettings[index];
    var id = account.getPropertyAsString("id");
    
    var answer = confirm("Konto '" + id + "', den Guthabenverlauf und\n"+
                         "die interne Gespr\u00E4chsliste l\u00F6schen?");
    
    if (answer) {
      // keep deleted account in a separate list (so the user can recover it)
      this._deletedAccounts.push(account);
      
      if (index < this._accountSettings.length-1)
        this._accountSettings[index] = this._accountSettings.pop();
      else
        this._accountSettings.pop();
      
      this._sortAndCompress();
      this._updateAccountList();
      this._switchAccount(aIndex);
      this._updatePreview();
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "options._deleteAccount",
                                "Deleted account '" + id + "'",
                                false);
    }
    else {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "options._deleteAccount",
                                "Cancelled dialog",
                                false);
    }
  },
  
  /****************/
  
  _updateMouseActionLists: function(aProvider, aRateplan) {
    var menulist1 = document.getElementById("options-ldest");
    var menulist2 = document.getElementById("options-mdest");
    
    // get current values
    var value1 = menulist1.selectedItem.value;
    var value2 = menulist1.selectedItem.value;
    
    if (typeof(value1) == "string" && value1.match(/\d+/)) {
      menulist1.selectedIndex = 0;
      value1 = Number(value1);
    }
    if (typeof(value2) == "string" && value2.match(/\d+/)) {
      menulist2.selectedIndex = 0;
      value2 = Number(value2);
    }
    
    // clean menu from provider-specific entries
    while (menulist1.removeItemAt(4)) { null; }
    while (menulist2.removeItemAt(4)) { null; }
    
    if (aProvider) {
      var links = this._managerCenter.getQuickLinkNames(aProvider, aRateplan);
      var index = 0;
      var n;
      while (links && links.hasMore()) {
        n = links.getNext();
        menulist1.appendItem(n, "" + index);
        menulist2.appendItem(n, "" + index);
        // we might want to check if index==value1/2 to reselect the menu item (TODO)
        index++;
        if (this._utils.isThunderbird)    /* only first entry for Thunderbird */
          break;
      }
    }
  },
  
  /****************/
  
  _changeManager: function(aNewProvider) {
    var menulist, value, check;
    
    // reload rate plans and reset selected rate plan
    this._updateRatePlanList(aNewProvider);
    menulist = document.getElementById("account-rateplan");
    menulist.selectedIndex = 0;
    this._selectedAccount.setPropertyAsInt("_rateplannum", 0);
    this._selectedAccount.setPropertyAsString("rateplan",
      menulist.selectedItem.value);
    
    this._changeRateplan(menulist.selectedItem.value);
    
    value = this._selectedAccount.getPropertyAsString("lclick-dest");
    menulist = document.getElementById("options-ldest");
    menulist.selectedIndex = 0;
    while (menulist.selectedItem) {
      if (menulist.selectedItem.value == value)
        break;
      else
        menulist.selectedIndex++;
    }
    if (!menulist.selectedItem) {
      menulist.selectedIndex = 0;
      value = menulist.selectedItem.value;
      this._selectedAccount.setPropertyAsString("lclick-dest", value);
    }
    document.getElementById("options-lurl").disabled = (value != "user");
    
    value = this._selectedAccount.getPropertyAsString("mclick-dest");
    menulist = document.getElementById("options-mdest");
    menulist.selectedIndex = 0;
    while (menulist.selectedItem) {
      if (menulist.selectedItem.value == value)
        break;
      else
        menulist.selectedIndex++;
    }
    if (!menulist.selectedItem) {
      menulist.selectedIndex = 0;
      value = menulist.selectedItem.value;
      this._selectedAccount.setPropertyAsString("mclick-dest", value);
    }
    document.getElementById("options-murl").disabled = (value != "user");
    
    // change the regular expr. for a valid user name and reevaluate it
    var regexpStr = this._managerCenter.getProviderProperties(aNewProvider).
                      getPropertyAsString("validUserRegExp");
    this._validUserRegExp = new RegExp(regexpStr);
    value = document.getElementById("account-login").value;
    if (!value.match(this._validUserRegExp))
      document.getElementById("account-login-label").setAttribute("style", "color: red;");
    else
      document.getElementById("account-login-label").removeAttribute("style");
  },
  
  /****************/
  
  _changeRateplan: function(aNewRateplan) {
    var account = this._selectedAccount;
    var provider = account.getPropertyAsString("provider");
    
    // reload quick-link menus and reset selected items
    this._updateMouseActionLists(provider, aNewRateplan);
    
    var prop = this._managerCenter.getRatePlanProperties(provider, aNewRateplan);
    if (prop) {
      if (prop.getPropertyAsBool("userResetPeriod")) {
        document.getElementById("tab-period").removeAttribute("hidden");
        document.getElementById("period-auto").disabled =
          !prop.getPropertyAsBool("autoResetPeriod");
      }
      else
        document.getElementById("tab-period").setAttribute("hidden", "true");
    }
    else
      document.getElementById("tab-period").setAttribute("hidden", "true");
  },
  
  /*********************************************************************/
  
  doInput: function(event) {
    var target = event.target;
    var id = target.id;
    var value = target.value;
    var changed = true;
    
    if ((id == null) || (id.length == 0))
      return true;
    
    switch (id) {
      // first handle the text boxes
      case "account-login":
        this._selectedAccount.setPropertyAsString("user", value);
        if (!value.match(this._validUserRegExp))
          document.getElementById("account-login-label").setAttribute("style", "color: red;");
        else
          document.getElementById("account-login-label").removeAttribute("style");
        break;
      
      case "account-password":
        //index = this._iconOrder[this._selectedAccount.getPropertyAsInt("position")];
        this._selectedAccount.setPropertyAsString("_password", value);
        break;
      
      case "account-initial":
        this._selectedAccount.setPropertyAsString("initial", value);
        if (this._selectedAccount.getPropertyAsBool("enabled"))
          this._selectedPanel.setAttribute("char", value);
        break;
      
      case "options-lurl":
        this._selectedAccount.setPropertyAsString("lclick-url", value);
        break;
      
      case "options-murl":
        this._selectedAccount.setPropertyAsString("mclick-url", value);
        break;
      
      default:
        //alert(id);
        changed = false;
        break;
    }
    
    if (changed)
      this._settingsChanged = true;
    
    return true;
  },
  
  /*********************************************************************/
  
  doCommand: function(event) {
    var target = event.target;
    var id = target.id;
    var value = target.value;
    //var index;
    var changed = true;
    
    if (target.nodeName == "menuitem") {
      while (target && (target.nodeName != "menulist"))
        target = target.parentNode;
      if (target)
        id = target.id;
    }
    else if (target.nodeName == "radio") {
      while (target && (target.nodeName != "radiogroup"))
        target = target.parentNode;
      if (target)
        id = target.id;
    }
    
    if ((id == null) || (id.length == 0))
      return true;
    
    switch (id) {
      // general settings
      case "options-nohide":
        this._globalSettings.setPropertyAsBool("autohide-all", !target.checked);
        break;
      
      case "options-divider":
        this._globalSettings.setPropertyAsBool("show-divider", target.checked);
        this._updatePreview();
        break;
      
      case "options-whatsnew":
        this._globalSettings.setPropertyAsBool("show-whatsnew", target.checked);
        break;
      
      // account settings
      case "account-list":
        this._switchAccount(target.selectedIndex);
        this._updatePreviewSelect();
        changed = false;
        break;
      
      case "account-add":
        this._addAccount();
        break;
      
      case "account-del":
        this._deleteAccount(document.getElementById("account-list").selectedIndex);
        break;
      
      case "account-rename":
        this._renameAccount(document.getElementById("account-list").selectedIndex);
        break;
      
      case "account-enabled":
        this._selectedAccount.setPropertyAsBool("enabled", target.checked);
        this._sortAndCompress();
        this._updatePreview();
        break;
      
      case "account-provider":
        if (this._selectedAccount.getPropertyAsString("provider") != value) {
          this._selectedAccount.setPropertyAsString("provider", value);
          this._changeManager(value);
        }
        break;
      
      case "account-rateplan":
        if (this._selectedAccount.getPropertyAsString("rateplan") != value) {
          this._selectedAccount.setPropertyAsString("rateplan", value);
          this._selectedAccount.setPropertyAsInt("_rateplannum", target.selectedIndex);
          this._changeRateplan(value);
        }
        break;
      
      case "account-pw-handling":
        this._selectedAccount.setPropertyAsString("pw-handling", value);
        if (value == "store")
          document.getElementById("account-password").removeAttribute("disabled");
        else
          document.getElementById("account-password").setAttribute("disabled", "true");
        break;
      
      case "account-icon":
        this._selectedAccount.setPropertyAsString("icon-theme", value);
        if (this._settings.getIconProperties(value).getPropertyAsBool("animated")) {
          document.getElementById("account-animate").removeAttribute("disabled");
          if (this._selectedAccount.getPropertyAsBool("animate"))
            document.getElementById("account-animate").checked = true;
        }
        else {
          document.getElementById("account-animate").setAttribute("disabled", "true");
          document.getElementById("account-animate").checked = false;
        }
        var acautoshow = document.getElementById("account-autoshow");
        if (this._settings.getIconProperties(value).getPropertyAsBool("hidden")) {
          acautoshow.selectedIndex = 0;
          document.getElementById("account-autoshow-no").setAttribute("disabled", "true");
          document.getElementById("account-autoshow-auto").setAttribute("disabled", "true");
        }
        else {
          var autoshow = this._selectedAccount.getPropertyAsString("show-label");
          acautoshow.selectedIndex =
            (autoshow == "no") ? 1 :
            (autoshow == "autoShow") ? 2 : 0;
          autoshow = acautoshow.selectedItem.value;
          this._selectedAccount.setPropertyAsString("show-label", autoshow);
          document.getElementById("account-autoshow-no").removeAttribute("disabled");
          document.getElementById("account-autoshow-auto").removeAttribute("disabled");
        }
        this._updatePreview();
        break;
      
      case "account-animate":
        this._selectedAccount.setPropertyAsBool("animate", target.checked);
        if (target.checked &&
            this._selectedAccount.getPropertyAsBool("enabled")) {
          // if the checkbox is checked, give the icon a kick!
          this._selectedPanel.animate = true;
          this._selectedPanel.animate = false;
        }
        break;
      
      case "account-autoshow":
        this._selectedAccount.setPropertyAsString("show-label", value);
        if (value == "autoShow")
          value = "autoShowHide";
        this._selectedPanel.showlabel = value;
        break;
      
      case "account-showage":
        this._selectedAccount.setPropertyAsBool("show-age", target.checked);
        break;
      
      case "icon-move-left":
        this._moveIconLeft();
        break;
      
      case "icon-move-right":
        this._moveIconRight();
        break;
      
      // mouse
      case "options-ldest":
        this._selectedAccount.setPropertyAsString("lclick-dest", value);
        if (value == "user")          // user-defined URL
          document.getElementById("options-lurl").removeAttribute("disabled");
        else
          document.getElementById("options-lurl").disabled = true;
        break;
      
      case "options-ltab":
        this._selectedAccount.setPropertyAsBool("lclick-tab", target.checked);
        break;
      
      case "options-mdest":
        this._selectedAccount.setPropertyAsString("mclick-dest", value);
        if (value == "user")          // user-defined URL
          document.getElementById("options-murl").removeAttribute("disabled");
        else
          document.getElementById("options-murl").disabled = true;
        break;
      
      case "options-mtab":
        this._selectedAccount.setPropertyAsBool("mclick-tab", target.checked);
        break;
      
      case "options-linksmenu":
        this._globalSettings.setPropertyAsBool("linksmenu-tab", target.checked);
        break;
      
      case "options-loglevel":
        this._globalSettings.setPropertyAsInt("loglevel", target.selectedIndex);
        break;
      
      case "options-showlog":
        this._utils.openFileInBrowser(this._utils.logFile.path, true);
        changed = false;
        break;
      
      case "options-emptylog":
        this._utils.emptyLogFile();
        document.getElementById("options-logsize").value = 
          this._utils.convertFileSize(this._utils.logFileSize);
        changed = false;
        break;
      
      case "period-reset":
        var mday_menu = document.getElementById("period-mday");
        if (target.selectedIndex == 1) {
          mday_menu.removeAttribute("disabled");
          this._selectedAccount.setPropertyAsInt("period-reset",
                                                 mday_menu.selectedIndex + 1);
        }
        else {
          mday_menu.disabled = true;
          this._selectedAccount.setPropertyAsString("period-reset", value);
        }
        break;
        
      case "period-mday":
        if (document.getElementById("period-reset").selectedIndex == 1)
          this._selectedAccount.setPropertyAsInt("period-reset",
                                                 target.selectedIndex + 1);
        break;
      
      default:
        //alert(id);
        changed = false;
        break;
    }
    
    if (changed)
      this._settingsChanged = true;
    
    return true;
  },
  
  /****************/
  
  _moveIconRight: function() {
    var current_pos = this._selectedAccount.getPropertyAsInt("position");
    //var current_index = this._iconOrder[current_pos];
    if ((current_pos + 1) < this._iconOrder.length) {
      var next_index = this._iconOrder[current_pos + 1];
      var next_account = this._accountSettings[next_index];
      if (next_account.getPropertyAsBool("enabled")) {
        // flip positions
        this._selectedAccount.setPropertyAsInt("position", current_pos + 1);
        next_account.setPropertyAsInt("position", current_pos);
        this._sortAndCompress();
        this._updatePreview();
      }
    }
  },
  
  /****************/
  
  _moveIconLeft: function() {
    var current_pos = this._selectedAccount.getPropertyAsInt("position");
    //var current_index = this._iconOrder[current_pos];
    if (current_pos > 0) {
      var prev_index = this._iconOrder[current_pos - 1];
      var prev_account = this._accountSettings[prev_index];
      if (prev_account.getPropertyAsBool("enabled")) {
        // flip positions
        this._selectedAccount.setPropertyAsInt("position", current_pos - 1);
        prev_account.setPropertyAsInt("position", current_pos);
        this._sortAndCompress();
        this._updatePreview();
      }
    }
  },
  
  /****************/
  
  // update the shown interval while slider is dragged
  _updateInterval: function(event) {
    if (event.attrName == "curpos") {
      var newint = Number(event.newValue) + 4;
      document.getElementById("options-interval").setAttribute("value", newint);
      gOpWindow._settingsChanged = true;
    }
  }
  
  
  
};


