/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2005,2006
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

// package constants
const DISPLAY_NAME   = "CheckYesss";
const NAME           = "checkyesss";
const UUID           = "{dbbdd4ec-1601-4ca9-87f3-98730c9fc8eb}";
const AUTHOR         = "Thomas Reitmayr";
const KEY            = "/" + NAME;
// should have been this, but now it's set once and for ever...
//const KEY            = "/" + AUTHOR + "/" + NAME;
const VERSION        = "0.6.24";
const LOCALE_LIST    = [ "de-DE" ];
const JAR_FILE       = NAME + ".jar";
const PREFS_FILE     = "defaults/preferences/" + NAME + ".js";
const AUTOREG_FILE   = "defaults/.autoreg";
const XPCOM_DIREC    = "components/";
const XPCOM_NAMES_JS = ["cyCheckyesss",
                        "cyCheckyesssUtil",
                        "cyCheckyesssZipWriter",
                        "cyCheckyesssSettings",
                        "cyCheckyesssContainer",
                        "cyCheckyesssCenter",
                        "cyCheckyesssWeb",
                        "cyCheckyesssSipgate",
                        "cyCheckyesssSipgateDE",
                        "cyCheckyesssVoipStation",
                        "cyCheckyesssSimyo"];
const XPCOM_NAMES_XPT = ["cyICheckyesss",
                         "cyICheckyesssUtil",
                         "cyICheckyesssZipWriter",
                         "cyICheckyesssSettings",
                         "cyICheckyesssContainer",
                         "cyICheckyesssCenter",
                         "cyICheckyesssWeb"];

const CONTENT_FOLDER = "content/";
const SKIN_FOLDER = "skin/classic/";

const WHERE_MESSAGE = DISPLAY_NAME + " fuer den aktuellen Benutzer (OK) oder " +
                      "alle Benutzer (Cancel/Abbruch) installieren?";

var err = null;

function checkNoError(aMessage) {
  err = getLastError();
  if (err != SUCCESS) {
    var errstr;
    switch (err) {
      case ACCESS_DENIED:  errstr = "ACCESS_DENIED"; break;
      case USER_CANCELLED: errstr = "USER_CANCELLED"; break;
      case READ_ONLY:      errstr = "READ_ONLY"; break;
      case INSUFFICIENT_DISK_SPACE: errstr = "INSUFFICIENT_DISK_SPACE"; break;
      case CHROME_REGISTRY_ERROR:   errstr = "CHROME_REGISTRY_ERROR"; break;
      default: errstr = err; break;
    }
    alert(aMessage + "\nCode: " + errstr);
    cancelInstall(err);
  }
  return err;
}

function doInstallation() {
  // begin the install
  initInstall(NAME, KEY, VERSION);
  if (checkNoError("Installation konnte nicht gestartet werden") != SUCCESS) { return; }

  var where, mainDir;
  var inProfile = confirm(WHERE_MESSAGE);
  if (inProfile) {
    mainDir = getFolder("Profile", "chrome");
    where = PROFILE_CHROME;
  }
  else {
    mainDir = getFolder("chrome");
    where = DELAYED_CHROME;
  }

  //var mainDir = getFolder("Profile", "extensions/" + UUID);
  //var chromeDir = getFolder(mainDir, "chrome");
  var chromeDir = mainDir;

  addFile(KEY, VERSION, "chrome/"+JAR_FILE, chromeDir, null);
  if (checkNoError(JAR_FILE + " konnte nicht angelegt werden,\n"+
                   "ev. fehlen die erforderlichen Rechte") != SUCCESS) { return; }
  
  // Add the defaults folder
  var defaultDir = getFolder(mainDir, "defaults");
  addDirectory(KEY, VERSION, "defaults", defaultDir, null);
  if (checkNoError("Defaults-Verzeichnis konnte nicht angelegt werden,\n"+
                   "ev. fehlen die erforderlichen Rechte") != SUCCESS) { return; }
  defaultDir = getFolder(getFolder("Program", "defaults"), "pref");
  addFile(KEY, VERSION, PREFS_FILE, defaultDir, null);
  if (checkNoError("Einstellungen konnten nicht angelegt werden,\n"+
                   "ev. fehlen die erforderlichen Rechte") != SUCCESS) { return; }
  
  //hack to make sure we register the component
  var pgmDir = getFolder("Program");
  addFile(KEY, VERSION, AUTOREG_FILE, pgmDir, null);
  if (checkNoError(AUTOREG_FILE + "-Datei konnte nicht angelegt werden") != SUCCESS) { return; }
  
  // Register the chrome URLs
  registerChrome(Install.CONTENT | where, getFolder(chromeDir, JAR_FILE), CONTENT_FOLDER);
  if (checkNoError(CONTENT_FOLDER + " konnte nicht registriert werden") != SUCCESS) { return; }
  registerChrome(Install.SKIN | where, getFolder(chromeDir, JAR_FILE), SKIN_FOLDER);
  if (checkNoError(SKIN_FOLDER + " konnte nicht registriert werden") != SUCCESS) { return; }

  for (var x = 0; x < LOCALE_LIST.length; x++) {
    registerChrome(Install.LOCALE | where, getFolder(chromeDir, JAR_FILE), "locale/"+LOCALE_LIST[x]+"/");
    if (checkNoError(LOCALE_LIST[x] + " konnte nicht registriert werden") != SUCCESS) { return; }
  }

  var compDir = getFolder("Components");
  for (x = 0; x < XPCOM_NAMES_JS.length; x++) {
    addFile(KEY, VERSION, XPCOM_DIREC+XPCOM_NAMES_JS[x]+".js", compDir, null);
    if (checkNoError(XPCOM_NAMES_JS[x] + ".js konnte nicht angelegt werden") != SUCCESS) { return; }
  }
  for (x = 0; x < XPCOM_NAMES_XPT.length; x++) {
    addFile(KEY, VERSION, XPCOM_DIREC+XPCOM_NAMES_XPT[x]+".xpt", compDir, null);
    if (checkNoError(XPCOM_NAMES_XPT[x] + ".xpt konnte nicht angelegt werden") != SUCCESS) { return; }
  }

  // Now install..
  performInstall();
  
  err = getLastError();
  if (err == SUCCESS) {
    alert(DISPLAY_NAME + " " + VERSION +
          " erfolgreich installiert!\nBitte den Browser neu starten.");
  }
  else {
    if (err == REBOOT_NEEDED) {
      checkNoError("Installation fehlgeschlagen!\n"+
                   "Vor dem naechsten Versuch den Browser neu starten\n"+
                   "oder "+DISPLAY_NAME+" zuerst deinstallieren");
    }
    else {
      checkNoError("Installation fehlgeschlagen!");
    }
  }
}


doInstallation();
